#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""csiread example: view the contents of a packet

Note: If the csi file is stored in HDD instead of SSD, it will take
      a little longer to read it for the first time.
"""

from timeit import default_timer

import csiread


def stringify(array, space=':'):
    return space.join([hex(element)[2:].zfill(2) for element in array])


def intel(csifile, index, ntxnum=2):
    """csitool"""
    print("="*40+"[intel]")
    members = [s for s in dir(csiread.Intel) if not s.startswith("__") and callable(getattr(csiread.Intel, s))]
    print("Methods: \n", members)

    print('Time:')
    last = default_timer()
    csidata = csiread.Intel(csifile, ntxnum=ntxnum, pl_size=10, if_report=False)
    csidata.read()
    print(" read                ", default_timer() - last, "s")

    last = default_timer()
    total_rss = csidata.get_total_rss()
    print(" get_total_rss       ", default_timer() - last, "s")

    last = default_timer()
    scaled_csi = csidata.get_scaled_csi()
    print(" get_scaled_csi      ", default_timer() - last, "s")

    last = default_timer()
    scaled_csi_sm = csidata.get_scaled_csi_sm()
    # scaled_csi_sm = csidata.apply_sm(scaled_csi)
    print(" get_scaled_csi_sm   ", default_timer() - last, "s")

    # Setting inplace to True may be dangerous but more efficient.
    temp = csiread.Intel(csifile, ntxnum=ntxnum, if_report=False)
    temp.read()
    last = default_timer()
    _ = temp.get_scaled_csi(inplace=True)       # _ is temp.csi == True
    print(" get_scaled_csi(T)   ", default_timer() - last, "s")

    temp = csiread.Intel(csifile, ntxnum=ntxnum, if_report=False)
    temp.read()
    last = default_timer()
    _ = temp.get_scaled_csi_sm(inplace=True)    # _ is temp.csi == True
    print(" get_scaled_csi_sm(T)", default_timer() - last, "s")

    print('-'*40)
    print("%dth packet: " % index)
    print(" file                ", csidata.file)
    print(" count               ", csidata.count)
    print(" timestamp_low       ", csidata.timestamp_low[index])
    print(" bfee_count          ", csidata.bfee_count[index])
    print(" Nrx                 ", csidata.Nrx[index])
    print(" Ntx                 ", csidata.Ntx[index])
    print(" rssi_a              ", csidata.rssi_a[index])
    print(" rssi_b              ", csidata.rssi_b[index])
    print(" rssi_c              ", csidata.rssi_c[index])
    print(" noise               ", csidata.noise[index])
    print(" agc                 ", csidata.agc[index])
    print(" perm                ", csidata.perm[index])
    print(" rate                ", csidata.rate[index])
    print(" csi                 ", csidata.csi[index].shape)
    print(" total_rss           ", total_rss[index])
    print(" scaled_csi.shape    ", scaled_csi[index].shape)
    print(" scaled_csi_sm.shape ", scaled_csi_sm[index].shape)
    if csidata.fc.size > index:
        print(" fc                  ", csidata.fc[index])
        print(" dur                 ", csidata.dur[index])
        print(" addr_src            ", stringify(csidata.addr_src[index], ":"))
        print(" addr_des            ", stringify(csidata.addr_des[index], ":"))
        print(" addr_bssid          ", stringify(csidata.addr_bssid[index], ":"))
        print(" seq                 ", csidata.seq[index])
        print(" payload             ", stringify(csidata.payload[index], " "))
    # print("a limitation: csidata.rate.size %d, csidata.rate.base.size %d" % (csidata.rate.size, csidata.rate.base.size))
    # print("help: \n", csidata.__doc__)


def atheros(csifile, index, ntxnum=2):
    """atheros"""
    print("="*40+"[atheros]")

    members = [s for s in dir(csiread.Atheros) if not s.startswith("__") and callable(getattr(csiread.Atheros, s))]
    print("Methods: \n", members)

    print('Time:')
    last = default_timer()
    csidata = csiread.Atheros(csifile, ntxnum=ntxnum, pl_size=10, if_report=False)
    csidata.read()
    print(" read                ", default_timer() - last, "s")

    print('-'*40)
    print("%dth packet: " % index)
    print(" file                ", csidata.file)
    print(" count               ", csidata.count)
    print(" timestamp           ", csidata.timestamp[index])
    print(" csi_len             ", csidata.csi_len[index])
    print(" tx_channel          ", csidata.tx_channel[index])
    print(" err_info            ", csidata.err_info[index])
    print(" noise_floor         ", csidata.noise_floor[index])
    print(" Rate                ", csidata.Rate[index])
    print(" bandWidth           ", csidata.bandWidth[index])
    print(" num_tones           ", csidata.num_tones[index])
    print(" nr                  ", csidata.nr[index])
    print(" nc                  ", csidata.nc[index])
    print(" rssi                ", csidata.rssi[index])
    print(" rssi_1              ", csidata.rssi_1[index])
    print(" rssi_2              ", csidata.rssi_2[index])
    print(" rssi_3              ", csidata.rssi_3[index])
    print(" payload_len         ", csidata.payload_len[index])
    print(" csi                 ", csidata.csi[index].shape)
    print(" payload             ", " ".join([hex(per)[2:].zfill(2) for per in csidata.payload[index]]))
    # print("help: \n", csidata.__doc__)


def nexmon(csifile, index, chip, bw):
    """nexmon"""
    print("="*40+"[nexmon]")

    if csiread.__version__ < '1.3.5':
        print("csiread.Nexmon: version >= 1.3.5 is required");return

    members = [s for s in dir(csiread.Nexmon) if not s.startswith("__") and callable(getattr(csiread.Nexmon, s))]
    print("Methods: \n", members)

    print('Time:')
    last = default_timer()
    csidata = csiread.Nexmon(csifile, chip, bw, if_report=False)
    csidata.read()
    print(" read                ", default_timer() - last, "s")

    print('-'*40)
    print("%dth packet: " % index)
    print(" file                ", csidata.file)
    print(" count               ", csidata.count)
    print(" chip                ", csidata.chip)
    print(" bw                  ", csidata.bw)
    print(" nano                ", csidata.nano)
    print(" sec                 ", csidata.sec[index])
    print(" usec                ", csidata.usec[index])
    print(" caplen              ", csidata.caplen[index])
    print(" wirelen             ", csidata.wirelen[index])
    print(" magic               ", hex(csidata.magic[index]))
    print(" src_addr            ", stringify(csidata.src_addr[index], ":"))
    print(" seq                 ", csidata.seq[index])
    print(" core                ", csidata.core[index])
    print(" spatial             ", csidata.spatial[index])
    print(" chan_spec           ", csidata.chan_spec[index])
    print(" chip_version        ", csidata.chip_version[index])
    print(" csi                 ", csidata.csi[index].shape)


if __name__ == "__main__":
    print("csiread.__version__: ", csiread.__version__)
    intel("../material/5300/dataset/sample_0x1_ap.dat", 10, ntxnum=2)
    intel("../material/5300/dataset/sample_0x5_64_3000.dat", 10, ntxnum=1)
    atheros("../material/atheros/dataset/ath_csi_1.dat", 10, ntxnum=2)
    nexmon("../material/nexmon/dataset/example.pcap", 0, '4358', 80)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""CSI userspace: Implement log_to_file in pure Python(Linux 802.11n CSI Tool)

Usage:
    sudo python3 csiuserspace.py

Note:
    I tested the example with the following code on Ubuntu18.04LTS without
    Intel 5300 NIC. The code hasn't been tested on the host equipped with
    Intel 5300 NIC.

    In Ubuntu18.04, CN_NETLINK_USERS = 11, but CN_NETLINK_USERS = 10 in the linux-80211n-csitool

    `csikernel.c` is based on[linux/samples/connector/cn_test.c]
    (https://github.com/torvalds/linux/blob/master/samples/connector/cn_test.c).

    ```bash
    make
    sudo insmod csikernel.ko
    dmesg
    ```

    Don't forget to remove csikernel after playing.

    ```bash
    sudo rmmod csikernel
    make clean
    ```

Ref: 
	1. [Kernel Connector](https://www.kernel.org/doc/html/latest/driver-api/connector.html)
	2. [connector.h](https://github.com/dhalperi/linux-80211n-csitool/blob/csitool-stable/include/linux/connector.h)
"""

import os
import socket
import struct


def log_to_file(csifile='csifile.dat'):
    """Implement log_to_file in Python"""
    f = open(csifile, 'wb')

    # show frequency
    count = 0
    SLOW_MSG_CNT = 1

    # /usr/include/linux/connector.h
    # #define CN_NETLINK_USERS		10	/* Highest index + 1 */
    CN_NETLINK_USERS = 10
    CN_IDX_IWLAGN = CN_NETLINK_USERS + 0xf
    # CN_VAL_IWLAGN = 0x1       # useless

    # /usr/include/linux/netlink.h
    # #define NETLINK_CONNECTOR	11
    # #define NETLINK_ADD_MEMBERSHIP		1
    socket.NETLINK_CONNECTOR = 11
    NETLINK_ADD_MEMBERSHIP = 1

    with socket.socket(socket.AF_NETLINK, socket.SOCK_DGRAM, socket.NETLINK_CONNECTOR) as s:
        # proc_addr
        s.bind((os.getpid(), CN_IDX_IWLAGN))

        # kern_addr, useless, pass

        # And subscribe to netlink group
        s.setsockopt(270, NETLINK_ADD_MEMBERSHIP, CN_IDX_IWLAGN)

        while True:
            ret = s.recv(4096)

            # /usr/include/linux/netlink.h
            # struct nlmsghdr {
            #     __u32		nlmsg_len;	/* Length of message including header */
            #     __u16		nlmsg_type;	/* Message content */
            #     __u16		nlmsg_flags;	/* Additional flags */
            #     __u32		nlmsg_seq;	/* Sequence number */
            #     __u32		nlmsg_pid;	/* Sending process port ID */
            # };

            nlmsg_len, nlmsg_type, nlmsg_flags, nlmsg_seq, nlmsg_pid = struct.unpack("=LHHLL", ret[:16])

            # /usr/include/linux/connector.h
            # struct cb_id {
            #     __u32 idx;
            #     __u32 val;
            # };

            # struct cn_msg {
            #     struct cb_id id;

            #     __u32 seq;
            #     __u32 ack;

            #     __u16 len;		/* Length of the following data */
            #     __u16 flags;
            #     __u8 data[0];
            # };

            cnmsg_idx, cnmsg_val, cnmsg_seq, cnmsg_ack, cnmsg_len, cnmsg_flag = struct.unpack("=LLLLHH", ret[16:36])
            
            # linux-80211n-csitool: /drivers/net/wireless/iwlwifi/iwl-connector.c
            # /**
            #  * Sends the message over the kernel connector to a userspace program.
            #  */
            # void connector_send_msg(const u8 *data, const u32 size, const u8 code)
            # {
            #     struct cn_msg *m;
            #     u8 *buf;
            #     u32 payload_size;

            #     /* Payload + 1-byte "code" */
            #     payload_size = size + 1 + sizeof(struct cn_msg);
            #     m = kmalloc(payload_size, GFP_ATOMIC);
            #     if (m == NULL) {
            #         printk(KERN_ERR "%s: malloc failed\n", __func__);
            #         return;
            #     }
            #     buf = ((u8 *) m) + sizeof(struct cn_msg);

            #     /* Set up message */
            #     memcpy(&m->id, &connector_id, sizeof(struct cb_id));
            #     m->seq = 0;
            #     m->len = size + 1;
            #     buf[0] = code;
            #     memcpy(&buf[1], data, size);

            #     /* Enqueue message -- may free on failure */
            #     connector_enqueue_msg(m);

            #     return;
            # }

            cnmsg_data = ret[36:]

            if count % SLOW_MSG_CNT == 0:
                print("received %d bytes: id: %d val: %d seq: %d clen: %d" % (cnmsg_len, cnmsg_idx, cnmsg_val, cnmsg_seq, cnmsg_len))
                # print("data: %s", bytes.decode(cnmsg_data))

            l2 = struct.pack('!H', cnmsg_len)
            f.write(l2)
            ret = f.write(cnmsg_data)

            if count % 100 == 0:
                print('wrote %d bytes [msgcnt=%u]' % (ret, count))
            count += 1
    f.close()


if __name__ == "__main__":
    log_to_file(csifile="csiuserspace.dat")



"""A fast channel state information parser for Intel, Atheros and Nexmon."""

from libc.stdio cimport (fopen, fread, fclose, fseek, ftell, printf, FILE,
                         SEEK_END, SEEK_SET, SEEK_CUR)
from libc.stdint cimport (uint16_t, int16_t, uint32_t, int32_t, uint8_t,
                          int8_t, uint64_t)
from libc.math cimport pi
import os
import struct

import numpy as np
cimport numpy as np
cimport cython


__version__ = "1.3.6"
__all__ = ['Intel', 'Atheros', 'Nexmon']


cdef class Intel:
    """Parse CSI obtained using 'Linux 802.11n CSI Tool'.

    Args:
        file (str or None): CSI data file. If ``str``, ``read`` and ``readstp``
            methods are allowed. If ``None``, ``seek`` and ``pmsg`` methods are
            allowed.
        nrxnum (int, optional): Number of receive antennas. Default: 3
        ntxnum (int, optional): Number of transmit antennas. Default: 2
        pl_size (int, optional): The size of payload to be used. Default: 0
        if_report (bool, optional): Report the parsed result. Default: ``True``
        bufsize (int, optional): The maximum amount of packets to be parsed.
            If ``0`` and file is ``str``, all packets will be parsed. If ``0``
            and file is ``None``, this parameter is ignored by `pmsg` method.
            Default: 0

    Attributes:
        file (str, readonly): CSI data file
        count (int, readonly): Count of 0xbb packets parsed
        timestamp_low (ndarray): The low 32 bits of the NIC's 1 MHz clock. It
            wraps about every 4300 seconds, or 72 minutes.
        bfee_count (ndarray): The count of the total number of beamforming
            measurements that have been recorded by the driver and sent to
            userspace. The netlink channel between the kernel and userspace is
            lossy, so these can be used to detect measurements that were
            dropped in this pipe.
        Nrx (ndarray): The number of antennas used to receive the packet.
        Ntx (ndarray): The number of space/time streams transmitted.
        rssi_a (ndarray): RSSI measured by the receiving NIC at the input to
            antenna port A. This measurement is made during the packet preamble.
            This value is in dB relative to an internal reference.
        rssi_b (ndarray): See ``rssi_a``
        rssi_c (ndarray): See ``rssi_a``
        noise (ndarray): Noise
        agc (ndarray): Automatic Gain Control (AGC) setting in dB
        perm (ndarray): Tell us how the NIC permuted the signals from the 3
            receive antennas into the 3 RF chains that process the measurements.
        rate (ndarray): The rate at which the packet was sent, in the same
            format as the ``rate_n_flags``.
        csi (ndarray): The CSI itself, normalized to an internal reference.
            It is a Count×30×Nrx×Ntx 4-D matrix where the second dimension is
            across 30 subcarriers in the OFDM channel. For a 20 MHz-wide
            channel, these correspond to about half the OFDM subcarriers, and
            for a 40 MHz-wide channel, this is about one in every 4 subcarriers.
        stp (ndarray): World timestamp recorded by the modified ``log_to_file``.
        fc (ndarray): Frame control
        dur (ndarray): Duration
        addr_des (ndarray): Destination MAC address
        addr_src (ndarray): Source MAC address
        addr_bssid (ndarray): BSSID MAC address
        seq (ndarray): Serial number of packet
        payload (ndarray): MAC frame to be used

    Examples:

        >>> csifile = "../material/5300/dataset/sample_0x1_ap.dat"
        >>> csidata = csiread.Intel(csifile, nrxnum=3, ntxnum=2, pl_size=10)
        >>> csidata.read()
        >>> csi = csidata.get_scaled_csi()
        >>> print(csidata.csi.shape)

    References:
        1. `Linux 802.11n CSI Tool <https://dhalperi.github.io/linux-80211n-csitool/>`_
        2. `linux-80211n-csitool-supplementary <https://github.com/dhalperi/linux-80211n-csitool-supplementary>`_
        3. `Linux 802.11n CSI Tool-FAQ <https://dhalperi.github.io/linux-80211n-csitool/faq.html>`_
    """
    cdef readonly str file
    cdef readonly int count

    cdef public np.ndarray timestamp_low
    cdef public np.ndarray bfee_count
    cdef public np.ndarray Nrx
    cdef public np.ndarray Ntx
    cdef public np.ndarray rssi_a
    cdef public np.ndarray rssi_b
    cdef public np.ndarray rssi_c
    cdef public np.ndarray noise
    cdef public np.ndarray agc
    cdef public np.ndarray perm
    cdef public np.ndarray rate
    cdef public np.ndarray csi
    cdef public np.ndarray stp

    cdef public np.ndarray fc
    cdef public np.ndarray dur
    cdef public np.ndarray addr_des
    cdef public np.ndarray addr_src
    cdef public np.ndarray addr_bssid
    cdef public np.ndarray seq
    cdef public np.ndarray payload

    cdef np.ndarray buf_timestamp_low
    cdef np.ndarray buf_bfee_count
    cdef np.ndarray buf_Nrx
    cdef np.ndarray buf_Ntx
    cdef np.ndarray buf_rssi_a
    cdef np.ndarray buf_rssi_b
    cdef np.ndarray buf_rssi_c
    cdef np.ndarray buf_noise
    cdef np.ndarray buf_agc
    cdef np.ndarray buf_perm
    cdef np.ndarray buf_rate
    cdef np.ndarray buf_csi
    cdef np.ndarray buf_stp

    cdef np.ndarray buf_fc
    cdef np.ndarray buf_dur
    cdef np.ndarray buf_addr_des
    cdef np.ndarray buf_addr_src
    cdef np.ndarray buf_addr_bssid
    cdef np.ndarray buf_seq
    cdef np.ndarray buf_payload

    cdef int nrxnum
    cdef int ntxnum
    cdef int pl_size
    cdef bint if_report

    def __cinit__(self, file, nrxnum=3, ntxnum=2, pl_size=0, if_report=True,
                  bufsize=0, *argv, **kw):
        self.file = file
        self.nrxnum = nrxnum
        self.ntxnum = ntxnum
        self.pl_size = pl_size
        self.if_report = if_report

        if bufsize == 0:
            if file is None:
                self.count = 1
                pk_num = 1
            else:
                lens = os.path.getsize(file)
                pk_num = lens // (35 + 60 * 1 * 1)
        else:
            pk_num = bufsize

        btype = np.int_

        self.buf_timestamp_low = np.zeros([pk_num], dtype=np.uint32)
        self.buf_bfee_count = np.zeros([pk_num], dtype=btype)
        self.buf_Nrx = np.zeros([pk_num], dtype=btype)
        self.buf_Ntx = np.zeros([pk_num], dtype=btype)
        self.buf_rssi_a = np.zeros([pk_num], dtype=btype)
        self.buf_rssi_b = np.zeros([pk_num], dtype=btype)
        self.buf_rssi_c = np.zeros([pk_num], dtype=btype)
        self.buf_noise = np.zeros([pk_num], dtype=btype)
        self.buf_agc = np.zeros([pk_num], dtype=btype)
        self.buf_perm = np.zeros([pk_num, 3], dtype=btype)
        self.buf_rate = np.zeros([pk_num], dtype=btype)
        self.buf_csi = np.zeros([pk_num, 30, self.nrxnum, self.ntxnum],
                                dtype=np.complex_)

        self.buf_fc = np.zeros([pk_num], dtype=btype)
        self.buf_dur = np.zeros([pk_num], dtype=btype)
        self.buf_addr_des = np.zeros([pk_num, 6], dtype=btype)
        self.buf_addr_src = np.zeros([pk_num, 6], dtype=btype)
        self.buf_addr_bssid = np.zeros([pk_num, 6], dtype=btype)
        self.buf_seq = np.zeros([pk_num], dtype=btype)
        self.buf_payload = np.zeros([pk_num, self.pl_size], dtype=np.uint8)

    def __init__(self, file, nrxnum=3, ntxnum=2, pl_size=0, if_report=True,
                 bufsize=0):
        pass

    def __getitem__(self, index):
        ret = {
            "timestamp_low": self.timestamp_low[index],
            "bfee_count": self.bfee_count[index],
            "Nrx": self.Nrx[index],
            "Ntx": self.Ntx[index],
            "rssi_a": self.rssi_a[index],
            "rssi_b": self.rssi_b[index],
            "rssi_c": self.rssi_c[index],
            "noise": self.noise[index],
            "agc": self.agc[index],
            "perm": self.perm[index],
            "rate": self.rate[index],
            "csi": self.csi[index]
        }
        return ret

    cpdef read(self):
        """Parse data if 0xbb and 0xc1 packets

        Examples:

            >>> csifile = "../material/5300/dataset/sample_0x1_ap.dat"
            >>> csidata = csiread.Intel(csifile)
            >>> csidata.read()
        """
        self.seek(self.file, 0, 0)

    cpdef seek(self, file, long pos, long num):
        """Read packets from a specific position

        This method allows us to read different parts of different files
        randomly. It could be useful in Machine Learning. However, it could be
        very slow when reading files in HDD for the first time. For this case,
        it is better to do a pre-read with ``read()`` first.

        Args:
            file (str): CSI data file.
            pos (int): Position of file descriptor corresponding to the packet.
                Currently, it must be returned by the function in
                ``example/csiseek.py``.
            num (int): Number of packets to be read. ``num <= bufsize`` must be
                true. If ``0``, all packets after ``pos`` will be read.

        Examples:

            >>> csifile = "../material/5300/dataset/sample_0x1_ap.dat"
            >>> csidata = csiread.Intel(None, bufsize=16)
            >>> for i in range(10):
            >>>     csidata.seek(csifile, 0, i+1)
            >>>     print(csidata.csi.shape)
        """
        cdef FILE *f

        tempfile = file.encode(encoding="utf-8")
        cdef char *datafile = tempfile

        f = fopen(datafile, "rb")
        if f is NULL:
            printf("Open failed!\n")
            fclose(f)
            return -1

        fseek(f, 0, SEEK_END)
        cdef long lens = ftell(f)
        fseek(f, pos, SEEK_SET)

        cdef np.uint32_t[:] buf_timestamp_low_mem = self.buf_timestamp_low
        cdef np.int_t[:] buf_bfee_count_mem = self.buf_bfee_count
        cdef np.int_t[:] buf_Nrx_mem = self.buf_Nrx
        cdef np.int_t[:] buf_Ntx_mem = self.buf_Ntx
        cdef np.int_t[:] buf_rssi_a_mem = self.buf_rssi_a
        cdef np.int_t[:] buf_rssi_b_mem = self.buf_rssi_b
        cdef np.int_t[:] buf_rssi_c_mem = self.buf_rssi_c
        cdef np.int_t[:] buf_noise_mem = self.buf_noise
        cdef np.int_t[:] buf_agc_mem = self.buf_agc
        cdef np.int_t[:, :] buf_perm_mem = self.buf_perm
        cdef np.int_t[:] buf_rate_mem = self.buf_rate
        cdef np.complex128_t[:, :, :, :] buf_csi_mem = self.buf_csi

        cdef np.int_t[:] buf_fc_mem = self.buf_fc
        cdef np.int_t[:] buf_dur_mem = self.buf_dur
        cdef np.int_t[:, :] buf_addr_des_mem = self.buf_addr_des
        cdef np.int_t[:, :] buf_addr_src_mem = self.buf_addr_src
        cdef np.int_t[:, :] buf_addr_bssid_mem = self.buf_addr_bssid
        cdef np.int_t[:] buf_seq_mem = self.buf_seq
        cdef np.uint8_t[:, :] buf_payload_mem = self.buf_payload

        cdef int count_0xbb = 0
        cdef int count_0xc1 = 0
        cdef unsigned short field_len
        cdef unsigned char code
        cdef unsigned char buf[1024]
        cdef unsigned char *payload
        cdef int l

        cdef int index, index_step
        cdef int i, j, k, g, perm_j
        cdef uint8_t remainder = 0
        cdef double a, b

        if num == 0:
            num = lens

        while pos < (lens-3):
            l = fread(&buf, sizeof(unsigned char), 3, f)
            field_len = buf[1] + (buf[0] << 8)
            code = buf[2]

            if code == 0xbb:
                l = fread(buf, sizeof(unsigned char), field_len - 1, f)
                if l != (field_len - 1):
                    break  # finished

                buf_timestamp_low_mem[count_0xbb] = cu32l(buf[0], buf[1],
                                                          buf[2], buf[3])
                buf_bfee_count_mem[count_0xbb] = cu16l(buf[4], buf[5])
                buf_Nrx_mem[count_0xbb] = buf[8]
                buf_Ntx_mem[count_0xbb] = buf[9]
                buf_rssi_a_mem[count_0xbb] = buf[10]
                buf_rssi_b_mem[count_0xbb] = buf[11]
                buf_rssi_c_mem[count_0xbb] = buf[12]
                buf_noise_mem[count_0xbb] = <int8_t>buf[13]
                buf_agc_mem[count_0xbb] = buf[14]
                buf_rate_mem[count_0xbb] = cu16l(buf[18], buf[19])

                buf_perm_mem[count_0xbb, 0] = (buf[15] & 0x3)
                buf_perm_mem[count_0xbb, 1] = ((buf[15] >> 2) & 0x3)
                buf_perm_mem[count_0xbb, 2] = ((buf[15] >> 4) & 0x3)

                if buf[8] > self.nrxnum:
                    fclose(f)
                    raise ValueError("nrxnum=%d is too small!\n" % self.nrxnum)
                if buf[9] > self.ntxnum:
                    fclose(f)
                    raise ValueError("ntxnum=%d is too small!\n" % self.ntxnum)
                if buf[16] | (buf[17] << 8) != 60 * buf[8] * buf[9] + 12:
                    fclose(f)
                    raise Exception("Wrong beamforming matrix size"
                                    ", %dth packet is broken!" % count_0xbb)

                payload = &buf[20]
                index = 0
                for i in range(30):
                    index = index + 3
                    remainder = index & 0x7
                    for j in range(buf[8]):
                        with cython.boundscheck(False):
                            perm_j = buf_perm_mem[count_0xbb, j]
                        for k in range(buf[9]):
                            index_step = index >> 3
                            a = ccsi(payload[index_step + 0],
                                     payload[index_step + 1], remainder)
                            b = ccsi(payload[index_step + 1],
                                     payload[index_step + 2], remainder)

                            set_csi_mem(buf_csi_mem, count_0xbb, i, perm_j, k,
                                        a, b)
                            index += 16
                count_0xbb += 1

            elif code == 0xc1:
                l = fread(buf, sizeof(unsigned char), field_len - 1, f)
                if l != (field_len - 1):
                    break  # finished

                buf_fc_mem[count_0xc1] = cu16l(buf[0], buf[1])
                buf_dur_mem[count_0xc1] = cu16l(buf[2], buf[3])

                for g in range(6):
                    buf_addr_des_mem[count_0xc1, g] = buf[4+g]
                    buf_addr_src_mem[count_0xc1, g] = buf[10+g]
                    buf_addr_bssid_mem[count_0xc1, g] = buf[16+g]

                buf_seq_mem[count_0xc1] = cu16l(buf[22], buf[23])

                for g in range(min(self.pl_size, field_len - 1)):
                    buf_payload_mem[count_0xc1, g] = buf[g]

                count_0xc1 += 1

            else:
                fseek(f, field_len - 1, SEEK_CUR)
            pos += (field_len + 2)
            if count_0xbb >= num:
                break

        fclose(f)

        if self.if_report:
            self.__report(count_0xbb, count_0xc1)

        del buf_timestamp_low_mem
        del buf_bfee_count_mem
        del buf_Nrx_mem
        del buf_Ntx_mem
        del buf_rssi_a_mem
        del buf_rssi_b_mem
        del buf_rssi_c_mem
        del buf_noise_mem
        del buf_agc_mem
        del buf_perm_mem
        del buf_rate_mem
        del buf_csi_mem

        del buf_fc_mem
        del buf_dur_mem
        del buf_addr_des_mem
        del buf_addr_src_mem
        del buf_addr_bssid_mem
        del buf_seq_mem
        del buf_payload_mem

        self.timestamp_low = self.buf_timestamp_low[:count_0xbb]
        self.bfee_count = self.buf_bfee_count[:count_0xbb]
        self.Nrx = self.buf_Nrx[:count_0xbb]
        self.Ntx = self.buf_Ntx[:count_0xbb]
        self.rssi_a = self.buf_rssi_a[:count_0xbb]
        self.rssi_b = self.buf_rssi_b[:count_0xbb]
        self.rssi_c = self.buf_rssi_c[:count_0xbb]
        self.noise = self.buf_noise[:count_0xbb]
        self.agc = self.buf_agc[:count_0xbb]
        self.perm = self.buf_perm[:count_0xbb, :]
        self.rate = self.buf_rate[:count_0xbb]
        self.csi = self.buf_csi[:count_0xbb, :, :, :]
        self.count = count_0xbb

        self.fc = self.buf_fc[:count_0xc1]
        self.dur = self.buf_dur[:count_0xc1]
        self.addr_des = self.buf_addr_des[:count_0xc1]
        self.addr_src = self.buf_addr_src[:count_0xc1]
        self.addr_bssid = self.buf_addr_bssid[:count_0xc1]
        self.seq = self.buf_seq[:count_0xc1]
        self.payload = self.buf_payload[:count_0xc1]

    cpdef pmsg(self, unsigned char *data):
        """Parse message in real time

        Args:
            data (bytes): A bytes object representing the data received by udp
                socket
        Returns:
            int: The status code. If ``0xbb`` and ``0xc1``, parse message
                successfully. Otherwise, the ``data`` is not a CSI packet.
        
        Examples:

            >>> import socket
            >>> import csiread
            >>> 
            >>> csidata = csiread.Intel(None)
            >>> with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            >>>     s.bind(('127.0.0.1', 10011))
            >>>     while True:
            >>>         data, address_src = s.recvfrom(4096)
            >>>         code = csidata.pmsg(data)
            >>>         if code == 0xbb:
            >>>             print(csidata.csi.shape)
        """
        cdef np.uint32_t[:] buf_timestamp_low_mem = self.buf_timestamp_low
        cdef np.int_t[:] buf_bfee_count_mem = self.buf_bfee_count
        cdef np.int_t[:] buf_Nrx_mem = self.buf_Nrx
        cdef np.int_t[:] buf_Ntx_mem = self.buf_Ntx
        cdef np.int_t[:] buf_rssi_a_mem = self.buf_rssi_a
        cdef np.int_t[:] buf_rssi_b_mem = self.buf_rssi_b
        cdef np.int_t[:] buf_rssi_c_mem = self.buf_rssi_c
        cdef np.int_t[:] buf_noise_mem = self.buf_noise
        cdef np.int_t[:] buf_agc_mem = self.buf_agc
        cdef np.int_t[:, :] buf_perm_mem = self.buf_perm
        cdef np.int_t[:] buf_rate_mem = self.buf_rate
        cdef np.complex128_t[:, :, :, :] buf_csi_mem = self.buf_csi

        cdef np.int_t[:] buf_fc_mem = self.buf_fc
        cdef np.int_t[:] buf_dur_mem = self.buf_dur
        cdef np.int_t[:, :] buf_addr_des_mem = self.buf_addr_des
        cdef np.int_t[:, :] buf_addr_src_mem = self.buf_addr_src
        cdef np.int_t[:, :] buf_addr_bssid_mem = self.buf_addr_bssid
        cdef np.int_t[:] buf_seq_mem = self.buf_seq
        cdef np.uint8_t[:, :] buf_payload_mem = self.buf_payload

        cdef unsigned char code
        cdef unsigned char *buf
        cdef unsigned char *payload

        cdef int index, index_step
        cdef int i, j, k, g, perm_j
        cdef uint8_t remainder = 0
        cdef double a, b

        code = data[0]
        buf = data
        buf += 1

        if code == 0xbb:
            buf_timestamp_low_mem[0] = cu32l(buf[0], buf[1], buf[2], buf[3])
            buf_bfee_count_mem[0] = cu16l(buf[4], buf[5])
            buf_Nrx_mem[0] = buf[8]
            buf_Ntx_mem[0] = buf[9]
            buf_rssi_a_mem[0] = buf[10]
            buf_rssi_b_mem[0] = buf[11]
            buf_rssi_c_mem[0] = buf[12]
            buf_noise_mem[0] = <int8_t>buf[13]
            buf_agc_mem[0] = buf[14]
            buf_rate_mem[0] = cu16l(buf[18], buf[19])

            buf_perm_mem[0, 0] = (buf[15] & 0x3)
            buf_perm_mem[0, 1] = ((buf[15] >> 2) & 0x3)
            buf_perm_mem[0, 2] = ((buf[15] >> 4) & 0x3)

            if buf[8] > self.nrxnum:
                raise ValueError("nrxnum=%d is too small!\n" % self.nrxnum)
            if buf[9] > self.ntxnum:
                raise ValueError("ntxnum=%d is too small!\n" % self.ntxnum)
            if buf[16] | (buf[17] << 8) != 60 * buf[8] * buf[9] + 12:
                printf("Wrong beamforming matrix size, the packet is broken!\n")
                return code

            payload = &buf[20]
            index = 0
            for i in range(30):
                index = index + 3
                remainder = index & 0x7
                for j in range(buf[8]):
                    with cython.boundscheck(False):
                        perm_j = buf_perm_mem[0, j]
                    for k in range(buf[9]):
                        index_step = index >> 3
                        a = ccsi(payload[index_step + 0],
                                 payload[index_step + 1], remainder)
                        b = ccsi(payload[index_step + 1],
                                 payload[index_step + 2], remainder)

                        set_csi_mem(buf_csi_mem, 0, i, perm_j, k, a, b)
                        index += 16
        if code == 0xc1:
            buf_fc_mem[0] = cu16l(buf[0], buf[1])
            buf_dur_mem[0] = cu16l(buf[2], buf[3])

            for g in range(6):
                buf_addr_des_mem[0, g] = buf[4+g]
                buf_addr_src_mem[0, g] = buf[10+g]
                buf_addr_bssid_mem[0, g] = buf[16+g]

            buf_seq_mem[0] = cu16l(buf[22], buf[23])

            for g in range(min(self.pl_size, len(data) - 1)):
                buf_payload_mem[0, g] = buf[g]

        del buf_timestamp_low_mem
        del buf_bfee_count_mem
        del buf_Nrx_mem
        del buf_Ntx_mem
        del buf_rssi_a_mem
        del buf_rssi_b_mem
        del buf_rssi_c_mem
        del buf_noise_mem
        del buf_agc_mem
        del buf_perm_mem
        del buf_rate_mem
        del buf_csi_mem

        del buf_fc_mem
        del buf_dur_mem
        del buf_addr_des_mem
        del buf_addr_src_mem
        del buf_addr_bssid_mem
        del buf_seq_mem
        del buf_payload_mem

        self.timestamp_low = self.buf_timestamp_low
        self.bfee_count = self.buf_bfee_count
        self.Nrx = self.buf_Nrx
        self.Ntx = self.buf_Ntx
        self.rssi_a = self.buf_rssi_a
        self.rssi_b = self.buf_rssi_b
        self.rssi_c = self.buf_rssi_c
        self.noise = self.buf_noise
        self.agc = self.buf_agc
        self.perm = self.buf_perm
        self.rate = self.buf_rate
        self.csi = self.buf_csi

        self.fc = self.buf_fc
        self.dur = self.buf_dur
        self.addr_des = self.buf_addr_des
        self.addr_src = self.buf_addr_src
        self.addr_bssid = self.buf_addr_bssid
        self.seq = self.buf_seq
        self.payload = self.buf_payload

        return code

    def readstp(self, endian='little'):
        """Parse timestamp recorded by the modified ``log_to_file``

        ``file.dat`` and ``file.datstp`` must be in the same directory.

        Args:
            endian (str): The byte order of ``file.datstp``， it can be
                ``little`` and ``big``. Default: ``little``
        
        Returns:
            int: Timestamp of the first packet.

        Examples:

            >>> csifile = "../material/5300/dataset/sample_0x1_ap.dat"
            >>> csidata = csiread.Intel(csifile)
            >>> first_stp = csidata.readstp()
            >>> print(first_stp)
        """
        self.stp = read_stpfile(self.file + "stp", endian)
        return self.stp[0]

    def get_total_rss(self):
        """Calculate the Received Signal Strength[RSS] in dBm from CSI
        
        Examples:

            >>> csifile = "../material/5300/dataset/sample_0x1_ap.dat"
            >>> csidata = csiread.Intel(csifile)
            >>> csidata.read()
            >>> rssi = csidata.get_total_rss()
            >>> print(rssi.shape)
        """
        rssi_mag = np.zeros_like(self.rssi_a, dtype=float)
        rssi_mag += self.__dbinvs(self.rssi_a)
        rssi_mag += self.__dbinvs(self.rssi_b)
        rssi_mag += self.__dbinvs(self.rssi_c)
        ret = self.__db(rssi_mag) - 44 - self.agc
        return ret

    cpdef get_scaled_csi(self, inplace=False):
        """Convert CSI to channel matrix H

        Args:
            inplace (bool): Optionally do the operation in-place. Default: False

        Returns:
            ndarray: Channel matrix H
        
        Examples:

            >>> csifile = "../material/5300/dataset/sample_0x1_ap.dat"
            >>> csidata = csiread.Intel(csifile)
            >>> csidata.read()
            >>> scaled_csi = csidata.get_scaled_csi(False)
            >>> print(scaled_csi.shape)
            >>> print("scaled_csi is csidata.csi: ", scaled_csi is csidata.csi)
        """
        cdef int i, j
        cdef int flat = 30 * self.nrxnum * self.ntxnum
        cdef double constant2 = 2
        cdef double constant4_5 = np.power(10, 0.45)
        cdef double temp_sum

        csi = self.csi
        csi_pwr = np.empty(self.count)
        cdef np.float64_t[:] csi_pwr_mem = csi_pwr
        cdef np.complex128_t[:, :] csi_mem = csi.reshape(self.count, flat)
        for i in range(self.count):
            temp_sum = 0
            for j in range(flat):
                with cython.boundscheck(False):
                    temp_sum += (csi_mem[i, j].real * csi_mem[i, j].real + 
                                 csi_mem[i, j].imag * csi_mem[i, j].imag)
            csi_pwr_mem[i] = temp_sum
        del csi_pwr_mem
        del csi_mem
        rssi_pwr = self.__dbinv(self.get_total_rss())

        scale = rssi_pwr / (csi_pwr / 30)

        noise_db = self.noise
        thermal_noise_pwr = self.__dbinv(noise_db)
        thermal_noise_pwr[noise_db == -127] = self.__dbinv(-92)

        quant_error_pwr = scale * (self.Nrx * self.Ntx)
        total_noise_pwr = thermal_noise_pwr + quant_error_pwr

        cdef np.float64_t[:] total_noise_pwr_mem = total_noise_pwr
        cdef np.int_t[:] Ntx_mem = self.Ntx
        for i in range(self.count):
            if Ntx_mem[i] == 2:
                total_noise_pwr_mem[i] = total_noise_pwr_mem[i] / constant2
            if Ntx_mem[i] == 3:
                total_noise_pwr_mem[i] = total_noise_pwr_mem[i] / constant4_5
        del Ntx_mem
        del total_noise_pwr_mem
        
        temp = np.sqrt(scale / total_noise_pwr).reshape(-1, 1, 1, 1)

        if inplace:
            self.csi *= temp
            return self.csi
        else:
            return self.csi * temp

    def get_scaled_csi_sm(self, inplace=False):
        """Convert CSI to pure channel matrix H

        This version undoes Intel's spatial mapping to return the pure MIMO
        channel matrix H.

        Args:
            inplace (bool): Optionally do the operation in-place. Default: False
        
        Returns:
            ndarray: The pure MIMO channel matrix H.
        
        Examples:

            >>> csifile = "../material/5300/dataset/sample_0x1_ap.dat"
            >>> csidata = csiread.Intel(csifile)
            >>> csidata.read()
            >>> scaled_csi_sm = csidata.get_scaled_csi_sm(False)
            >>> print(scaled_csi.shape)
            >>> print("scaled_csi_sm is csidata.csi: ", scaled_csi_sm is csidata.csi)
        """
        if inplace:
            return self.__remove_sm(self.get_scaled_csi(True), True)
        else:
            return self.__remove_sm(self.get_scaled_csi(False), False)

    def apply_sm(self, scaled_csi):
        """Undo the input spatial mapping

        Args:
            scaled_csi (ndarray): Channel matrix H.

        Returns:
            ndarray: The pure MIMO channel matrix H.
        
        Examples:

            >>> csifile = "../material/5300/dataset/sample_0x1_ap.dat"
            >>> csidata = csiread.Intel(csifile)
            >>> csidata.read()
            >>> scaled_csi = csidata.get_scaled_csi()
            >>> scaled_csi_sm = csidata.apply_sm(scaled_csi)
            >>> print(scaled_csi_sm.shape)
        """
        return self.__remove_sm(scaled_csi)

    def __report(self, int count_0xbb, int count_0xc1):
        """Report parsed result."""
        if count_0xbb == 0:
            printf("connector_log=0x%x\n", 4)
            printf("%d 0xc1 packets parsed\n", count_0xc1)
        elif count_0xc1 == 0:
            printf("connector_log=0x%x\n", 1)
            printf("%d 0xbb packets parsed\n", count_0xbb)
        else:
            printf("connector_log=0x%x\n", 1 | 4)
            printf("%d 0xc1 packets parsed\n", count_0xc1)
            printf("%d 0xbb packets parsed\n", count_0xbb)

    cdef __dbinvs(self, x):
        """Convert from decibels specially"""
        ret = np.power(10, x / 10)
        ret[ret == 1] = 0
        return ret

    cdef __dbinv(self, x):
        """Convert from decibels"""
        ret = np.power(10, x / 10)
        return ret

    cdef __db(self, x):
        """Calculate decibels"""
        ret = 10 * np.log10(x)
        return ret

    cdef __remove_sm(self, scaled_csi, inplace=False):
        """Actually undo the input spatial mapping

        Args:
            scaled_csi (ndarray): Channel matrix H.
            inplace (bool): Optionally do the operation in-place. Default: False

        Returns:
            ndarray: The pure MIMO channel matrix H.
        """
        #  Conjugate Transpose
        sm_2_20 = np.array([[1,  1],
                            [1, -1]],
                           dtype=np.complex_) / np.sqrt(2)
        sm_2_40 = np.array([[ 1, -1j],
                            [-1j, 1]],
                           dtype=np.complex_) / np.sqrt(2)
        sm_3_20 = np.array([[-2*pi/16,      2*pi/(80/23), -2*pi/(80/13)],
                            [-2*pi/(80/33), 2*pi/(48/13),  2*pi/(240/37)],
                            [ 2*pi/(80/3),  2*pi/(240/13), 2*pi/(48/13)]],
                           dtype=np.complex_)
        sm_3_20 = np.power(np.e, -1j * sm_3_20) / np.sqrt(3)
        sm_3_40 = np.array([[-2*pi/16,      -2*pi/(80/37),    2*pi/(80/7)],
                            [-2*pi/(80/13), -2*pi/(48/11),   -2*pi/(240/83)],
                            [ 2*pi/(80/23), -2*pi/(240/107), -2*pi/(48/11)]],
                           dtype=np.complex_)
        sm_3_40 = np.power(np.e, -1j * sm_3_40) / np.sqrt(3)

        # Ntx is not a constant array
        if inplace:
            ret = scaled_csi
        else:
            ret = np.zeros([self.count, 30, self.nrxnum, self.ntxnum],
                           dtype=np.complex_)

        cdef int i, N, M, B
        cdef np.int_t[:] Ntx_mem = self.Ntx
        cdef np.int_t[:] Nrx_mem = self.Nrx
        cdef np.int_t[:] rate_mem = self.rate
        cdef np.complex128_t[:, :, :, :] scaled_csi_mem = scaled_csi
        cdef np.complex128_t[:, :, :, :] ret_mem = ret
        cdef np.complex128_t[:, :] sm_2_20_mem = sm_2_20
        cdef np.complex128_t[:, :] sm_2_40_mem = sm_2_40
        cdef np.complex128_t[:, :] sm_3_20_mem = sm_3_20
        cdef np.complex128_t[:, :] sm_3_40_mem = sm_3_40

        for i in range(self.count):
            M = Ntx_mem[i]
            N = Nrx_mem[i]
            B = (rate_mem[i] & 0x800) == 0x800
            if B:
                if M == 3:
                    intel_mm_o3(ret_mem[i], scaled_csi_mem[i], sm_3_40_mem,
                                N, M)
                elif M == 2:
                    intel_mm_o3(ret_mem[i], scaled_csi_mem[i], sm_2_40_mem,
                                N, M)
                else:
                    if inplace is False:
                        ret_mem[i] = scaled_csi_mem[i]
            else:
                if M == 3:
                    intel_mm_o3(ret_mem[i], scaled_csi_mem[i], sm_3_20_mem,
                                N, M)
                elif M == 2:
                    intel_mm_o3(ret_mem[i], scaled_csi_mem[i], sm_2_20_mem,
                                N, M)
                else:
                    if inplace is False:
                        ret_mem[i] = scaled_csi_mem[i]
        del Ntx_mem
        del Nrx_mem
        del rate_mem
        del scaled_csi_mem
        del ret_mem
        del sm_2_20_mem
        del sm_2_40_mem
        del sm_3_20_mem
        del sm_3_40_mem

        return ret


cdef class Atheros:
    """Parse CSI obtained using 'Atheros CSI Tool'.

    Args:
        file (str or None): CSI data file. If ``str``, ``read`` and ``readstp``
            methods are allowed. If ``None``, ``seek`` and ``pmsg`` methods are
            allowed.
        nrxnum (int, optional): Number of receive antennas. Default: 3
        ntxnum (int, optional): Number of transmit antennas. Default: 2
        pl_size (int, optional): The size of payload to be used. Default: 0
        tones (int, optional): The number of subcarrier. It can be 56 and 114.
            Default: 56
        if_report (bool, optional): Report the parsed result. Default: ``True``
        bufsize (int, optional): The maximum amount of packets to be parsed.
            If ``0`` and file is ``str``, all packets will be parsed. If ``0``
            and file is ``None``, this parameter is ignored by ``pmsg`` method.
            Default: 0

    Attributes:
        file (str, readonly): CSI data file
        count (int, readonly): Count of CSI packets parsed
        timestamp (ndarray): The time when packet is received, expressed in μs
        csi_len (ndarray): The csi data length in the received data buffer,
            expressed in bytes
        tx_channel (ndarray): The center frequency of the wireless channel,
            expressed in MHz
        err_info (ndarray): The phy error code, set to 0 if correctly received
        noise_floor (ndarray): The noise floor, expressed in dB. But it needs
            to be update and is set to 0 in current version.
        Rate (ndarray): The data rate of the received packet. Its value is a
            unsigned 8 bit integer number and the mapping between this value
            and the rate choice of 802.11 protocol
        bandWidth (ndarray): The channel bandwidth. It is 20MHz if set to 0 and
            40MHz if set to 1
        num_tones (ndarray): The number of subcarrier that used for data
            transmission.
        nr (ndarray): Number of receiving antenna
        nc (ndarray): Number of transmitting antenna
        rsssi (ndarray): The rssi of combination of all active chains
        rssi_1 (ndarray): The rssi of active chain 0
        rssi_2 (ndarray): The rssi of active chain 1
        rssi_3 (ndarray): The rssi of active chain 2
        payload_len (ndarray): The payload length of received packet, expressed
            in bytes.
        csi (ndarray): CSI
        payload (ndarray): MAC frame(MPDU) to be used

    Examples:

        >>> csifile = "../material/atheros/dataset/ath_csi_1.dat"
        >>> csidata = csiread.Atheros(csifile, nrxnum=3, ntxnum=2, pl_size=10, tones=56)
        >>> csidata.read(endian='little')
        >>> print(csidata.csi.shape)

    References:
        1. `Atheros CSI Tool <https://wands.sg/research/wifi/AtherosCSI/>`_
        2. `Atheros-CSI-Tool-UserSpace-APP <https://github.com/xieyaxiongfly/Atheros-CSI-Tool-UserSpace-APP>`_
        3. `Atheros CSI Tool User Guide <https://wands.sg/research/wifi/AtherosCSI/document/Atheros-CSI-Tool-User-Guide.pdf>`_
    """
    cdef readonly str file
    cdef readonly int count

    cdef public np.ndarray timestamp
    cdef public np.ndarray csi_len
    cdef public np.ndarray tx_channel
    cdef public np.ndarray err_info
    cdef public np.ndarray noise_floor
    cdef public np.ndarray Rate
    cdef public np.ndarray bandWidth
    cdef public np.ndarray num_tones
    cdef public np.ndarray nr
    cdef public np.ndarray nc
    cdef public np.ndarray rssi
    cdef public np.ndarray rssi_1
    cdef public np.ndarray rssi_2
    cdef public np.ndarray rssi_3
    cdef public np.ndarray payload_len
    cdef public np.ndarray csi
    cdef public np.ndarray payload

    cdef np.ndarray buf_timestamp
    cdef np.ndarray buf_csi_len
    cdef np.ndarray buf_tx_channel
    cdef np.ndarray buf_err_info
    cdef np.ndarray buf_noise_floor
    cdef np.ndarray buf_Rate
    cdef np.ndarray buf_bandWidth
    cdef np.ndarray buf_num_tones
    cdef np.ndarray buf_nr
    cdef np.ndarray buf_nc
    cdef np.ndarray buf_rssi
    cdef np.ndarray buf_rssi_1
    cdef np.ndarray buf_rssi_2
    cdef np.ndarray buf_rssi_3
    cdef np.ndarray buf_payload_len
    cdef np.ndarray buf_csi
    cdef np.ndarray buf_payload

    cdef int nrxnum
    cdef int ntxnum
    cdef int tones
    cdef int pl_size
    cdef bint if_report

    def __cinit__(self, file, nrxnum=3, ntxnum=2, pl_size=0, tones=56,
                  if_report=True, bufsize=0, *argv, **kw):
        self.file = file
        self.nrxnum = nrxnum
        self.ntxnum = ntxnum
        self.tones = tones
        self.pl_size = pl_size
        self.if_report = if_report

        if tones not in [56, 114]:
            raise ValueError("tones can only take 56 and 114!\n")

        if bufsize == 0:
            if file is None:
                self.count = 1
                pk_num = 1
            else:
                lens = os.path.getsize(file)
                pk_num = int(lens / 420)
        else:
            pk_num = bufsize

        btype = np.int_
        self.buf_timestamp = np.zeros([pk_num], dtype=np.uint64)
        self.buf_csi_len = np.zeros([pk_num], dtype=btype)
        self.buf_tx_channel = np.zeros([pk_num], dtype=btype)
        self.buf_err_info = np.zeros([pk_num], dtype=btype)
        self.buf_noise_floor = np.zeros([pk_num], dtype=btype)
        self.buf_Rate = np.zeros([pk_num], dtype=btype)
        self.buf_bandWidth = np.zeros([pk_num], dtype=btype)
        self.buf_num_tones = np.zeros([pk_num], dtype=btype)
        self.buf_nr = np.zeros([pk_num], dtype=btype)
        self.buf_nc = np.zeros([pk_num], dtype=btype)
        self.buf_rssi = np.zeros([pk_num], dtype=btype)
        self.buf_rssi_1 = np.zeros([pk_num], dtype=btype)
        self.buf_rssi_2 = np.zeros([pk_num], dtype=btype)
        self.buf_rssi_3 = np.zeros([pk_num], dtype=btype)
        self.buf_payload_len = np.zeros([pk_num], dtype=btype)
        self.buf_csi = np.zeros([pk_num, self.tones, self.nrxnum, self.ntxnum],
                                dtype=np.complex_)
        self.buf_payload = np.zeros([pk_num, self.pl_size], dtype=np.uint8)

    def __init__(self, file, nrxnum=3, ntxnum=2, pl_size=0, tones=56,
                 if_report=True, bufsize=0):
        pass

    def __getitem__(self, index):
        ret = {
            "timestamp": self.timestamp[index],
            "csi_len": self.csi_len[index],
            "tx_channel": self.tx_channel[index],
            "err_info": self.err_info[index],
            "noise_floor": self.noise_floor[index],
            "Rate": self.Rate[index],
            "bandWidth": self.bandWidth[index],
            "num_tones": self.num_tones[index],
            "nr": self.nr[index],
            "nc": self.nc[index],
            "rssi": self.rssi[index],
            "rssi_1": self.rssi_1[index],
            "rssi_2": self.rssi_2[index],
            "rssi_3": self.rssi_3[index],
            "payload_len": self.payload_len[index],
            "csi": self.csi[index],
            "payload": self.payload[index]
        }
        return ret

    cpdef read(self, endian='little'):
        """Parse data

        Args:
            endian (str): The byte order of ``file.dat``， it can be ``little``
                and ``big``. Default: ``little``
            
        Examples:

            >>> csifile = "../material/atheros/dataset/ath_csi_1.dat"
            >>> csidata = csiread.Atheros(csifile)
            >>> csidata.read()
        """
        self.seek(self.file, 0, 0, endian)

    cpdef seek(self, file, long pos, long num, endian='little'):
        """Read packets from a specific position

        This method allows us to read different parts of different files
        randomly. It could be useful in Machine Learning. However, it could be
        very slow when reading files in HDD for the first time. For this case,
        it is better to do a pre-read with ``read()`` first.

        Args:
            file (str): CSI data file.
            pos (int): Position of file descriptor corresponding to the packet.
                Currently, it must be returned by the function in
                `example/csiseek.py`.
            num (int): Number of packets to be read. ``num <= bufsize`` must be
                true. If ``0``, all packets after ``pos`` will be read.
            endian (str): The byte order of ``file.dat``， it can be ``little``
                and ``big``. Default: ``little``
        
        Examples:

            >>> csifile = "../material/atheros/dataset/ath_csi_1.dat"
            >>> csidata = csiread.Atheros(None, bufsize=16)
            >>> for i in range(10):
            >>>     csidata.seek(csifile, 0, i+1)
            >>>     print(csidata.csi.shape)
        """
        cdef FILE *f

        tempfile = file.encode(encoding="utf-8")
        cdef char *datafile = tempfile

        f = fopen(datafile, "rb")
        if f is NULL:
            printf("Open failed!\n")
            fclose(f)
            return -1

        fseek(f, 0, SEEK_END)
        cdef long lens = ftell(f)
        fseek(f, pos, SEEK_SET)

        cdef np.uint64_t[:] buf_timestamp_mem = self.buf_timestamp
        cdef np.int_t[:] buf_csi_len_mem = self.buf_csi_len
        cdef np.int_t[:] buf_tx_channel_mem = self.buf_tx_channel
        cdef np.int_t[:] buf_err_info_mem = self.buf_err_info
        cdef np.int_t[:] buf_noise_floor_mem = self.buf_noise_floor
        cdef np.int_t[:] buf_Rate_mem = self.buf_Rate
        cdef np.int_t[:] buf_bandWidth_mem = self.buf_bandWidth
        cdef np.int_t[:] buf_num_tones_mem = self.buf_num_tones
        cdef np.int_t[:] buf_nr_mem = self.buf_nr
        cdef np.int_t[:] buf_nc_mem = self.buf_nc
        cdef np.int_t[:] buf_rssi_mem = self.buf_rssi
        cdef np.int_t[:] buf_rssi_1_mem = self.buf_rssi_1
        cdef np.int_t[:] buf_rssi_2_mem = self.buf_rssi_2
        cdef np.int_t[:] buf_rssi_3_mem = self.buf_rssi_3
        cdef np.int_t[:] buf_payload_len_mem = self.buf_payload_len
        cdef np.complex128_t[:, :, :, :] buf_csi_mem = self.buf_csi
        cdef np.uint8_t[:, :] buf_payload_mem = self.buf_payload

        cdef int count = 0
        cdef int c_len, pl_len, pl_stop
        cdef int l, field_len
        cdef uint16_t (*ath_cu16)(uint8_t, uint8_t)
        cdef uint64_t (*ath_cu64)(uint64_t, uint64_t, uint64_t, uint64_t,
                                  uint64_t, uint64_t, uint64_t, uint64_t)

        if num == 0:
            num = lens

        cdef int bits_left, bitmask, idx, h_data, current_data
        cdef int k, nc_idx, nr_idx, imag, real, i
        cdef unsigned char buf[4096]
        cdef unsigned char csi_buf[4096]
        if endian == "little":
            ath_cu16 = cu16l
            ath_cu64 = cu64l
        elif endian == "big":
            ath_cu16 = cu16b
            ath_cu64 = cu64b
        else:
            fclose(f)
            raise ValueError("endian must be either 'little' or 'big'")

        while pos < (lens - 4):
            l = fread(&buf, sizeof(unsigned char), 2, f)
            field_len = ath_cu16(buf[0], buf[1])
            pos += 2
            if (pos + field_len) > lens:
                break

            l = fread(&buf, sizeof(unsigned char), 25, f)
            buf_timestamp_mem[count] = ath_cu64(buf[0], buf[1], buf[2], buf[3],
                                                buf[4], buf[5], buf[6], buf[7])
            buf_csi_len_mem[count] = ath_cu16(buf[8], buf[9])
            buf_tx_channel_mem[count] = ath_cu16(buf[10], buf[11])
            buf_payload_len_mem[count] = ath_cu16(buf[23], buf[24])
            buf_err_info_mem[count] = buf[12]
            buf_noise_floor_mem[count] = buf[13]
            buf_Rate_mem[count] = buf[14]
            buf_bandWidth_mem[count] = buf[15]
            buf_num_tones_mem[count] = buf[16]
            buf_nr_mem[count] = buf[17]
            buf_nc_mem[count] = buf[18]
            buf_rssi_mem[count] = buf[19]
            buf_rssi_1_mem[count] = buf[20]
            buf_rssi_2_mem[count] = buf[21]
            buf_rssi_3_mem[count] = buf[22]
            pos += 25

            if buf[17] > self.nrxnum:
                fclose(f)
                raise ValueError("nrxnum=%d is too small!\n" % self.nrxnum)
            if buf[18] > self.ntxnum:
                fclose(f)
                raise ValueError("ntxnum=%d is too small!\n" % self.ntxnum)

            c_len = buf_csi_len_mem[count]
            if c_len > 0:
                l = fread(&csi_buf, sizeof(unsigned char), c_len, f)
                bits_left = 16
                bitmask = (1 << 10) - 1

                idx = 0
                h_data = csi_buf[idx]
                idx += 1
                h_data += (csi_buf[idx] << 8)
                idx += 1
                current_data = h_data & ((1 << 16) - 1)

                for k in range(buf[16]):
                    for nr_idx in range(buf[17]):
                        for nc_idx in range(buf[18]):
                            # imag
                            if (bits_left - 10) < 0:
                                h_data = csi_buf[idx]
                                idx += 1
                                h_data += (csi_buf[idx] << 8)
                                idx += 1
                                current_data += h_data << bits_left
                                bits_left += 16
                            imag = current_data & bitmask
                            if imag & (1 << 9):
                                imag -= (1 << 10)

                            bits_left -= 10
                            current_data = current_data >> 10
                            # real
                            if (bits_left - 10) < 0:
                                h_data = csi_buf[idx]
                                idx += 1
                                h_data += (csi_buf[idx] << 8)
                                idx += 1
                                current_data += h_data << bits_left
                                bits_left += 16
                            real = current_data & bitmask
                            if real & (1 << 9):
                                real -= (1 << 10)

                            bits_left -= 10
                            current_data = current_data >> 10
                            # csi
                            set_csi_mem(buf_csi_mem, count, k, nr_idx, nc_idx,
                                        real, imag)
                pos += c_len

            pl_len = buf_payload_len_mem[count]
            pl_stop = min(pl_len, self.pl_size)
            if pl_len > 0:
                l = fread(&buf, sizeof(unsigned char), pl_len, f)
                for i in range(pl_stop):
                    buf_payload_mem[count, i] = buf[i]
                pos += pl_len

            # In matlab, read_log_file drops the last two packets, but here we 
            # keep them.
            count += 1
            if count >= num:
                break

        fclose(f)

        if self.if_report:
            self.__report(count)

        del buf_timestamp_mem
        del buf_csi_len_mem
        del buf_tx_channel_mem
        del buf_err_info_mem
        del buf_noise_floor_mem
        del buf_Rate_mem
        del buf_bandWidth_mem
        del buf_num_tones_mem
        del buf_nr_mem
        del buf_nc_mem
        del buf_rssi_mem
        del buf_rssi_1_mem
        del buf_rssi_2_mem
        del buf_rssi_3_mem
        del buf_payload_len_mem
        del buf_csi_mem
        del buf_payload_mem

        self.timestamp = self.buf_timestamp[:count]
        self.csi_len = self.buf_csi_len[:count]
        self.tx_channel = self.buf_tx_channel[:count]
        self.err_info = self.buf_err_info[:count]
        self.noise_floor = self.buf_noise_floor[:count]
        self.Rate = self.buf_Rate[:count]
        self.bandWidth = self.buf_bandWidth[:count]
        self.num_tones = self.buf_num_tones[:count]
        self.nr = self.buf_nr[:count]
        self.nc = self.buf_nc[:count]
        self.rssi = self.buf_rssi[:count]
        self.rssi_1 = self.buf_rssi_1[:count]
        self.rssi_2 = self.buf_rssi_2[:count]
        self.rssi_3 = self.buf_rssi_3[:count]
        self.payload_len = self.buf_payload_len[:count]
        self.csi = self.buf_csi[:count]
        self.payload = self.buf_payload[:count]
        self.count = count

    cpdef pmsg(self, unsigned char *data, endian='little'):
        """Parse message in real time

        Args:
            data (bytes): A bytes object representing the data received by udp
                socket
            endian (str): The byte order of ``file.dat``， it can be ``little``
                and ``big``. Default: ``little``

        Returns:
            int: The status code. If ``0xff00``, parse message successfully.
                Otherwise, the ``data`` is not a CSI packet.

        Examples:

            >>> import socket
            >>> import csiread
            >>> 
            >>> csidata = csiread.Atheros(None)
            >>> with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            >>>     s.bind(('127.0.0.1', 10011))
            >>>     while True:
            >>>         data, address_src = s.recvfrom(4096)
            >>>         code = csidata.pmsg(data)
            >>>         if code == 0xff00:
            >>>             print(csidata.csi.shape)
        """
        cdef np.uint64_t[:] buf_timestamp_mem = self.buf_timestamp
        cdef np.int_t[:] buf_csi_len_mem = self.buf_csi_len
        cdef np.int_t[:] buf_tx_channel_mem = self.buf_tx_channel
        cdef np.int_t[:] buf_err_info_mem = self.buf_err_info
        cdef np.int_t[:] buf_noise_floor_mem = self.buf_noise_floor
        cdef np.int_t[:] buf_Rate_mem = self.buf_Rate
        cdef np.int_t[:] buf_bandWidth_mem = self.buf_bandWidth
        cdef np.int_t[:] buf_num_tones_mem = self.buf_num_tones
        cdef np.int_t[:] buf_nr_mem = self.buf_nr
        cdef np.int_t[:] buf_nc_mem = self.buf_nc
        cdef np.int_t[:] buf_rssi_mem = self.buf_rssi
        cdef np.int_t[:] buf_rssi_1_mem = self.buf_rssi_1
        cdef np.int_t[:] buf_rssi_2_mem = self.buf_rssi_2
        cdef np.int_t[:] buf_rssi_3_mem = self.buf_rssi_3
        cdef np.int_t[:] buf_payload_len_mem = self.buf_payload_len
        cdef np.complex128_t[:, :, :, :] buf_csi_mem = self.buf_csi
        cdef np.uint8_t[:, :] buf_payload_mem = self.buf_payload

        cdef int count = 0
        cdef int c_len, pl_len, pl_stop

        cdef int bits_left, bitmask, idx, h_data, current_data
        cdef int k, nc_idx, nr_idx, imag, real, i
        cdef unsigned char *buf
        cdef unsigned char *csi_buf
        cdef uint16_t (*ath_cu16)(uint8_t, uint8_t)
        cdef uint64_t (*ath_cu64)(uint64_t, uint64_t, uint64_t, uint64_t,
                                  uint64_t, uint64_t, uint64_t, uint64_t)

        buf = data
        if endian == "little":
            ath_cu16 = cu16l
            ath_cu64 = cu64l
        elif endian == "big":
            ath_cu16 = cu16b
            ath_cu64 = cu64b
        else:
            raise ValueError("endian must be either 'little' or 'big'")
        buf_timestamp_mem[count] = ath_cu64(buf[0], buf[1], buf[2], buf[3],
                                            buf[4], buf[5], buf[6], buf[7])
        buf_csi_len_mem[count] = ath_cu16(buf[8], buf[9])
        buf_tx_channel_mem[count] = ath_cu16(buf[10], buf[11])
        buf_payload_len_mem[count] = ath_cu16(buf[23], buf[24])
        buf_err_info_mem[count] = buf[12]
        buf_noise_floor_mem[count] = buf[13]
        buf_Rate_mem[count] = buf[14]
        buf_bandWidth_mem[count] = buf[15]
        buf_num_tones_mem[count] = buf[16]
        buf_nr_mem[count] = buf[17]
        buf_nc_mem[count] = buf[18]
        buf_rssi_mem[count] = buf[19]
        buf_rssi_1_mem[count] = buf[20]
        buf_rssi_2_mem[count] = buf[21]
        buf_rssi_3_mem[count] = buf[22]

        if buf[17] > self.nrxnum:
            raise ValueError("nrxnum=%d is too small!\n" % self.nrxnum)
        if buf[18] > self.ntxnum:
            raise ValueError("ntxnum=%d is too small!\n" % self.ntxnum)

        c_len = buf_csi_len_mem[count]
        if c_len > 0:
            csi_buf = &buf[25]
            bits_left = 16
            bitmask = (1 << 10) - 1

            idx = 0
            h_data = csi_buf[idx]
            idx += 1
            h_data += (csi_buf[idx] << 8)
            idx += 1
            current_data = h_data & ((1 << 16) - 1)

            for k in range(buf[16]):
                for nr_idx in range(buf[17]):
                    for nc_idx in range(buf[18]):
                        # imag
                        if (bits_left - 10) < 0:
                            h_data = csi_buf[idx]
                            idx += 1
                            h_data += (csi_buf[idx] << 8)
                            idx += 1
                            current_data += h_data << bits_left
                            bits_left += 16
                        imag = current_data & bitmask
                        if imag & (1 << 9):
                            imag -= (1 << 10)

                        bits_left -= 10
                        current_data = current_data >> 10
                        # real
                        if (bits_left - 10) < 0:
                            h_data = csi_buf[idx]
                            idx += 1
                            h_data += (csi_buf[idx] << 8)
                            idx += 1
                            current_data += h_data << bits_left
                            bits_left += 16
                        real = current_data & bitmask
                        if real & (1 << 9):
                            real -= (1 << 10)

                        bits_left -= 10
                        current_data = current_data >> 10
                        # csi
                        set_csi_mem(buf_csi_mem, count, k, nr_idx, nc_idx,
                                    real, imag)

        pl_len = buf_payload_len_mem[count]
        pl_stop = min(pl_len, self.pl_size)
        if pl_len > 0:
            for i in range(pl_stop):
                buf_payload_mem[count, i] = buf[25+c_len+i]

        del buf_timestamp_mem
        del buf_csi_len_mem
        del buf_tx_channel_mem
        del buf_err_info_mem
        del buf_noise_floor_mem
        del buf_Rate_mem
        del buf_bandWidth_mem
        del buf_num_tones_mem
        del buf_nr_mem
        del buf_nc_mem
        del buf_rssi_mem
        del buf_rssi_1_mem
        del buf_rssi_2_mem
        del buf_rssi_3_mem
        del buf_payload_len_mem
        del buf_csi_mem
        del buf_payload_mem

        self.timestamp = self.buf_timestamp
        self.csi_len = self.buf_csi_len
        self.tx_channel = self.buf_tx_channel
        self.err_info = self.buf_err_info
        self.noise_floor = self.buf_noise_floor
        self.Rate = self.buf_Rate
        self.bandWidth = self.buf_bandWidth
        self.num_tones = self.buf_num_tones
        self.nr = self.buf_nr
        self.nc = self.buf_nc
        self.rssi = self.buf_rssi
        self.rssi_1 = self.buf_rssi_1
        self.rssi_2 = self.buf_rssi_2
        self.rssi_3 = self.buf_rssi_3
        self.payload_len = self.buf_payload_len
        self.csi = self.buf_csi
        self.payload = self.buf_payload

        return 0xff00

    def readstp(self, endian='little'):
        """Parse timestamp recorded by the modified ``recv_csi``

        ``file.dat`` and ``file.datstp`` must be in the same directory.

        Args:
            endian (str): The byte order of ``file.datstp``， it can be
                ``little`` and ``big``. Default: ``little``

        Examples:

            >>> csifile = "../material/atheros/dataset/ath_csi_1.dat"
            >>> csidata = csiread.Atheros(csifile)
            >>> first_stp = csidata.readstp()
            >>> print(first_stp)
        """
        self.stp = read_stpfile(self.file + "stp", endian)
        return self.stp[0]

    def __report(self, int count):
        """Report parsed result."""
        printf("%d packets parsed\n", count)


cdef class Nexmon:
    """Parse CSI obtained using 'nexmon_csi'.

    Args:
        file (str or None): CSI data file ``.pcap``. If ``str``, ``read``
            methods is allowed. If ``None``, ``seek`` and ``pmsg`` methods are
            allowed.
        chip (str): WiFi Chip, it can be '4339', '43455c0', '4358' and '4366c0'.
        bw (int): bandwidth, it can be 20, 40 and 80.
        if_report (bool, optional): Report the parsed result. Default: `True`
        bufsize (int, optional): The maximum amount of packets to be parsed. If
            ``0`` and file is ``str``, all packets will be parsed. If ``0`` and
            file is ``None``, this parameter is ignored by `pmsg` method.
            Default: 0
    
    Attributes:
        file (str, readonly): CSI data file
        count (int, readonly): Count of csi packets parsed
        chip (str, readonly): Chip type we set
        bw (int, readonly): Bandwidth we set
        nano (bool, readonly): nanosecond-resolution or not
        sec (ndarray): Time the packet was captured
        usec (ndarray): The microseconds when this packet was captured, as an
            offset to ``sec`` if ``nano`` is True. The nanoseconds when the
            packet was captured, as an offset to ``sec`` if ``nano`` is False.
        caplen (ndarray): The number of bytes of packet data actually captured
            and saved in the file
        wirelen (ndarray): The length of the packet as it appeared on the
            network when it was captured
        magic (ndarray): Four magic bytes ``0x11111111``
        src_addr (ndarray): Source MAC address 
        seq (ndarray): Sequence number of the Wi-Fi frame that triggered the
            collection of the CSI contained in packets
        core (ndarray): Core
        spatial (ndarray): Spatial stream 
        chan_spec (ndarray): (unknown)
        chip_version (ndarray): The chip version
        csi (ndarray): CSI

    Examples:

        >>> csifile = "../material/nexmon/dataset/example.pcap"
        >>> csidata = csiread.Nexmon(csifile, chip='4358', bw=80)
        >>> csidata.read()
        >>> print(csidata.csi.shape)

    References:
        1. `nexmon_csi <https://github.com/seemoo-lab/nexmon_csi>`_
        2. `rdpcap <https://github.com/secdev/scapy/blob/master/scapy/utils.py>`_
        3. `Libpcap File Format <https://wiki.wireshark.org/Development/LibpcapFileFormat>`_
    """
    cdef readonly str file
    cdef readonly int count
    cdef readonly str chip
    cdef readonly int bw
    cdef readonly bint nano

    cdef public np.ndarray sec
    cdef public np.ndarray usec
    cdef public np.ndarray caplen
    cdef public np.ndarray wirelen
    cdef public np.ndarray magic
    cdef public np.ndarray src_addr
    cdef public np.ndarray seq
    cdef public np.ndarray core
    cdef public np.ndarray spatial
    cdef public np.ndarray chan_spec
    cdef public np.ndarray chip_version
    cdef public np.ndarray csi

    cdef np.ndarray buf_sec
    cdef np.ndarray buf_usec
    cdef np.ndarray buf_caplen
    cdef np.ndarray buf_wirelen
    cdef np.ndarray buf_magic
    cdef np.ndarray buf_src_addr
    cdef np.ndarray buf_seq
    cdef np.ndarray buf_core
    cdef np.ndarray buf_spatial
    cdef np.ndarray buf_chan_spec
    cdef np.ndarray buf_chip_version
    cdef np.ndarray buf_csi

    cdef bint if_report
    cdef public int _autoscale

    def __cinit__(self, file, chip, bw, if_report=True, bufsize=0,
                  *argv, **kw):
        self.file = file
        self.chip = chip
        self.bw = bw
        self.if_report = if_report

        if bufsize == 0:
            if file is None:
                self.count = 1
                pk_num = 1
            else:
                pk_num = self.__get_count()
        else:
            pk_num = bufsize

        btype = np.int_
        self.buf_sec = np.zeros([pk_num], dtype=np.uint32)
        self.buf_usec = np.zeros([pk_num], dtype=np.uint32)
        self.buf_caplen = np.zeros([pk_num], dtype=btype)
        self.buf_wirelen = np.zeros([pk_num], dtype=btype)
        self.buf_magic = np.zeros([pk_num], dtype=btype)
        self.buf_src_addr = np.zeros([pk_num, 6], dtype=btype)
        self.buf_seq = np.zeros([pk_num], dtype=btype)
        self.buf_core = np.zeros([pk_num], dtype=btype)
        self.buf_spatial = np.zeros([pk_num], dtype=btype)
        self.buf_chan_spec = np.zeros([pk_num], dtype=btype)
        self.buf_chip_version = np.zeros([pk_num], dtype=btype)
        self.buf_csi = np.zeros([pk_num, int(self.bw * 3.2)],
                                dtype=np.complex_)
        self._autoscale = 1

    def __init__(self, file, chip, bw, if_report=True, bufsize=0):
        pass

    def __getitem__(self, index):
        ret = {
            "magic": self.magic[index],
            "src_addr": self.src_addr[index],
            "seq": self.seq[index],
            "core": self.core[index],
            "spatial": self.spatial[index],
            "chan_spec": self.chan_spec[index],
            "chip_version": self.chip_version[index],
            "csi": self.csi[index]
        }
        return ret

    cpdef read(self):
        """Parse data
        
        Examples:

            >>> csifile = "../material/nexmon/dataset/example.pcap"
            >>> csidata = csiread.Nexmon(csifile, chip='4358', bw=80)
            >>> csidata.read()
            >>> print(csidata.csi.shape)
        """
        self.seek(self.file, 24, 0)

    cpdef seek(self, file, long pos, long num):
        """Read packets from specific position

        This method allows us to read different parts of different files
        randomly. It could be useful in Machine Learning. However, it could be
        very slow when reading files in HDD for the first time. For this case,
        it is better to use `read()` for a pre-read first.

        Args:
            file (str): CSI data file ``.pcap``.
            pos (int): Position of file descriptor corresponding to the packet.
                Currently, it must be returned by the function in 
                ``example/csiseek.py``.
            num (int): Number of packets to be read. ``num <= bufsize`` must be
                true. If ``0``, all packets after ``pos`` will be read.

        Examples:

            >>> csifile = "../material/nexmon/dataset/example.pcap"
            >>> csidata = csiread.Nexmon(None, chip='4358', bw=80, bufsize=4)
            >>> for i in range(4):
            >>>     csidata.seek(csifile, 0, i+1)
            >>>     print(csidata.csi.shape)
        """
        cdef FILE *f

        tempfile = file.encode(encoding="utf-8")
        cdef char *datafile = tempfile

        f = fopen(datafile, "rb")
        if f is NULL:
            printf("Open failed!\n")
            fclose(f)
            return -1

        endian = self.__pcapheader(f)
        fseek(f, 0, SEEK_END)
        cdef long lens = ftell(f)
        fseek(f, pos, SEEK_SET)

        cdef np.uint32_t[:] buf_sec_mem = self.buf_sec
        cdef np.uint32_t[:] buf_usec_mem = self.buf_usec
        cdef np.int_t[:] buf_caplen_mem = self.buf_caplen
        cdef np.int_t[:] buf_wirelen_mem = self.buf_wirelen
        cdef np.int_t[:] buf_magic_mem = self.buf_magic
        cdef np.int_t[:, :] buf_src_addr_mem = self.buf_src_addr
        cdef np.int_t[:] buf_seq_mem = self.buf_seq
        cdef np.int_t[:] buf_core_mem = self.buf_core
        cdef np.int_t[:] buf_spatial_mem = self.buf_spatial
        cdef np.int_t[:] buf_chan_spec_mem = self.buf_chan_spec
        cdef np.int_t[:] buf_chip_version_mem = self.buf_chip_version
        cdef np.complex128_t[:, :] buf_csi_mem = self.buf_csi

        cdef int count = 0
        cdef unsigned char buf[4096]
        cdef int l, i
        cdef int nfft = <int>(self.bw * 3.2)
        cdef uint32_t caplen
        cdef bint flag
        cdef uint16_t (*nex_cu16)(uint8_t, uint8_t) 
        cdef uint32_t (*nex_cu32)(uint8_t, uint8_t, uint8_t, uint8_t)

        if num == 0:
            num = lens

        if endian == "little":
            nex_cu16 = cu16l
            nex_cu32 = cu32l
            flag = True
        else:
            nex_cu16 = cu16b
            nex_cu32 = cu32b
            flag = False

        while pos < (lens - 24):
            # global header
            l = fread(&buf, sizeof(unsigned char), 16, f)
            if l < 16:
                break
            caplen = nex_cu32(buf[8], buf[9], buf[10], buf[11])
            buf_sec_mem[count] = nex_cu32(buf[0], buf[1], buf[2], buf[3])
            buf_usec_mem[count] = nex_cu32(buf[4], buf[5], buf[6], buf[7])
            buf_caplen_mem[count] = caplen
            buf_wirelen_mem[count] = nex_cu32(buf[12], buf[13], buf[14],
                                              buf[15])
            pos += (16 + caplen)

            # we don't care about enth+ip+udp header
            l = fread(&buf, sizeof(unsigned char), 42, f)
            if buf[6:12] != b'NEXMON':
                fseek(f, caplen - 42, SEEK_CUR)
                continue

            # nexmon header
            l = fread(&buf, sizeof(unsigned char), 18, f)
            buf_magic_mem[count] = nex_cu32(buf[0], buf[1], buf[2], buf[3])
            for i in range(6):
                buf_src_addr_mem[count, i] = buf[4+i]
            buf_seq_mem[count] = nex_cu16(buf[10], buf[11])
            buf_core_mem[count] = nex_cu16(buf[12], buf[13]) & 0x7
            buf_spatial_mem[count] = (nex_cu16(buf[12], buf[13]) >> 3) & 0x7
            buf_chan_spec_mem[count] = nex_cu16(buf[14], buf[15])
            buf_chip_version_mem[count] = nex_cu16(buf[16], buf[17])

            # CSI
            l = fread(&buf, sizeof(unsigned char), caplen - 42 - 18, f)
            if self.chip == '4339' or self.chip == '43455c0':
                unpack_int16(buf, buf_csi_mem[count], nfft, flag)
            elif self.chip == '4358':
                unpack_float(buf, buf_csi_mem[count], nfft, 9, 5,
                             self._autoscale, flag)
            elif self.chip == '4366c0':
                unpack_float(buf, buf_csi_mem[count], nfft, 12, 6,
                             self._autoscale, flag)
            else:
                pass

            count += 1
            if count >= num:
                break
        fclose(f)
        self.count = count
        if self.if_report:
            printf("%d packets parsed\n", count)
        del buf_sec_mem
        del buf_usec_mem
        del buf_caplen_mem
        del buf_wirelen_mem
        del buf_magic_mem
        del buf_src_addr_mem
        del buf_seq_mem
        del buf_core_mem
        del buf_spatial_mem
        del buf_chan_spec_mem
        del buf_chip_version_mem
        del buf_csi_mem

        self.sec = self.buf_sec[:count]
        self.usec = self.buf_usec[:count]
        self.caplen = self.buf_caplen[:count]
        self.wirelen = self.buf_wirelen[:count]
        self.magic = self.buf_magic[:count]
        self.src_addr = self.buf_src_addr[:count]
        self.seq = self.buf_seq[:count]
        self.core = self.buf_core[:count]
        self.spatial = self.buf_spatial[:count]
        self.chan_spec = self.buf_chan_spec[:count]
        self.chip_version = self.buf_chip_version[:count]
        self.csi = self.buf_csi[:count]

    cpdef pmsg(self, unsigned char *data, endian='little'):
        """Parse message in real time

        Args:
            data (bytes): A bytes object representing the data received by raw
                socket
            endian (str): The byte order of ``file.dat``， it can be ``little``
                and ``big``. Default: ``little``

        Returns:
            int: The status code. If ``0xf100``, parse message successfully.
                Otherwise, the ``data`` is not a CSI packet.
        
        Examples:

            >>> import socket
            >>> import csiread
            >>> 
            >>> csidata = csiread.Nexmon(None, chip='4358', bw=80)
            >>> with socket.socket(socket.PF_PACKET, socket.SOCK_RAW, socket.htons(0x3)) as s:
            >>>     while True:
            >>>         data, address_src = s.recvfrom(4096)
            >>>         code = csidata.pmsg(data)
            >>>         if code == 0xf100:
            >>>             print(csidata.csi.shape)
        """
        cdef np.int_t[:] buf_magic_mem = self.buf_magic
        cdef np.int_t[:, :] buf_src_addr_mem = self.buf_src_addr
        cdef np.int_t[:] buf_seq_mem = self.buf_seq
        cdef np.int_t[:] buf_core_mem = self.buf_core
        cdef np.int_t[:] buf_spatial_mem = self.buf_spatial
        cdef np.int_t[:] buf_chan_spec_mem = self.buf_chan_spec
        cdef np.int_t[:] buf_chip_version_mem = self.buf_chip_version
        cdef np.complex128_t[:, :] buf_csi_mem = self.buf_csi

        cdef int count = 0
        cdef unsigned char *buf
        cdef int l, i
        cdef int nfft = <int>(self.bw * 3.2)
        cdef bint flag
        cdef uint16_t (*nex_cu16)(uint8_t, uint8_t) 
        cdef uint32_t (*nex_cu32)(uint8_t, uint8_t, uint8_t, uint8_t)

        buf = data

        if endian == "little":
            nex_cu16 = cu16l
            nex_cu32 = cu32l
            flag = True
        else:
            nex_cu16 = cu16b
            nex_cu32 = cu32b
            flag = False

        # we don't care about enth+ip+udp header
        if buf[6:12] != b'NEXMON':
            return

        # nexmon header
        buf_magic_mem[count] = nex_cu32(buf[42], buf[43], buf[44], buf[45])
        for i in range(6):
            buf_src_addr_mem[count, i] = buf[46+i]
        buf_seq_mem[count] = nex_cu16(buf[52], buf[53])
        buf_core_mem[count] = nex_cu16(buf[54], buf[55]) & 0x7
        buf_spatial_mem[count] = (nex_cu16(buf[54], buf[55]) >> 3) & 0x7
        buf_chan_spec_mem[count] = nex_cu16(buf[56], buf[57])
        buf_chip_version_mem[count] = nex_cu16(buf[58], buf[59])

        # CSI
        if self.chip == '4339' or self.chip == '3455c0':
            unpack_int16(&buf[60], buf_csi_mem[count], nfft, flag)
        elif self.chip == '4358':
            unpack_float(&buf[60], buf_csi_mem[count], nfft, 9, 5,
                         self._autoscale, flag)
        elif self.chip == '4366c0':
            unpack_float(&buf[60], buf_csi_mem[count], nfft, 12, 6,
                         self._autoscale, flag)
        else:
            pass

        del buf_magic_mem
        del buf_src_addr_mem
        del buf_seq_mem
        del buf_core_mem
        del buf_spatial_mem
        del buf_chan_spec_mem
        del buf_chip_version_mem
        del buf_csi_mem

        self.magic = self.buf_magic
        self.src_addr = self.buf_src_addr
        self.seq = self.buf_seq
        self.core = self.buf_core
        self.spatial = self.buf_spatial
        self.chan_spec = self.buf_chan_spec
        self.chip_version = self.buf_chip_version
        self.csi = self.buf_csi

        return 0xf100

    cdef __get_count(self):
        cdef FILE *f
        tempfile = self.file.encode(encoding="utf-8")
        cdef char *datafile = tempfile

        f = fopen(datafile, "rb")
        if f is NULL:
            printf("Open failed!\n")
            fclose(f)
            return -1

        cdef int count = 0
        cdef int l
        cdef uint32_t caplen
        cdef unsigned char buf[64]
        cdef uint32_t (*nex_cu32)(uint8_t, uint8_t, uint8_t, uint8_t)

        # pcap header: head: endian
        endian = self.__pcapheader(f)
        if endian == 'little':
            nex_cu32 = cu32l
        else:
            nex_cu32 = cu32b

        # count
        while True:
            l = fread(&buf, sizeof(unsigned char), 16+42, f)
            if l < 16:
                break
            if buf[22:28] == b'NEXMON':
                count += 1
            caplen = nex_cu32(buf[8], buf[9], buf[10], buf[11])
            fseek(f, caplen - 42, SEEK_CUR)
        fclose(f)
        return count

    cdef __pcapheader(self, FILE *f):
        cdef unsigned char buf[16]
        cdef int l

        l = fread(&buf, sizeof(unsigned char), 4, f)
        magic = buf[:4]
        if magic == b"\xa1\xb2\xc3\xd4":    # big endian
            endian = "big"
            self.nano = False
        elif magic == b"\xd4\xc3\xb2\xa1":  # little endian
            endian = "little"
            self.nano = False
        elif magic == b"\xa1\xb2\x3c\x4d":  # big endian, nanosecond
            endian = "big"
            self.nano = True
        elif magic == b"\x4d\x3c\xb2\xa1":  # little endian, nanosecond
            endian = "little"
            self.nano = True
        else:
            raise Exception("Not a pcap capture file (bad magic: %r)" % magic)

        fseek(f, 20, SEEK_CUR)
        return endian


cdef read_stpfile(stpfile, endian):
    lens = os.path.getsize(stpfile) // 8
    stp = np.zeros(lens)
    format_string = '<LL' if endian == 'little' else '>LL'
    cdef int i, a, b
    f = open(stpfile, "rb")
    for i in range(lens):
        a, b = struct.unpack(format_string, f.read(8))
        stp[i] = a + b / 1000000
    f.close()
    return stp


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void intel_mm_o3(np.complex128_t[:, :, :] ret_mem,
                      np.complex128_t[:, :, :] scaled_csi_mem,
                      np.complex128_t[:, :] sm, int N, int M):
    """Matrix multiplication of O^3
    
    The function uses a trick of GEMM. It can be faster by using OPENMP and
    BLAS, but needs more dependencies.
    """
    cdef int i, j
    cdef np.complex128_t sm00, sm01, sm02, sm10, sm11, sm12, sm20, sm21, sm22
    cdef np.complex128_t r0, r1, r2

    if M == 2:
        sm00 = sm[0, 0]
        sm01 = sm[0, 1]
        sm10 = sm[1, 0]
        sm11 = sm[1, 1]
        for i in range(30):
            for j in range(N):
                r0 = scaled_csi_mem[i, j, 0]
                r1 = scaled_csi_mem[i, j, 1]
                ret_mem[i, j, 0] = r0 * sm00 + r1 * sm10
                ret_mem[i, j, 1] = r0 * sm01 + r1 * sm11
    if M == 3:
        sm00 = sm[0, 0]
        sm01 = sm[0, 1]
        sm02 = sm[0, 2]
        sm10 = sm[1, 0]
        sm11 = sm[1, 1]
        sm12 = sm[1, 2]
        sm20 = sm[2, 0]
        sm21 = sm[2, 1]
        sm22 = sm[2, 2]
        for i in range(30):
            for j in range(N):
                r0 = scaled_csi_mem[i, j, 0]
                r1 = scaled_csi_mem[i, j, 1]
                r2 = scaled_csi_mem[i, j, 2]
                ret_mem[i, j, 0] = r0 * sm00 + r1 * sm10 + r2 * sm20
                ret_mem[i, j, 1] = r0 * sm01 + r1 * sm11 + r2 * sm21
                ret_mem[i, j, 2] = r0 * sm02 + r1 * sm12 + r2 * sm22


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void set_csi_mem(np.complex128_t[:, :, :, :] csi_mem, int count,
                             int s, int r, int t, double real, double imag):
    csi_mem[count, s, r, t].real = real
    csi_mem[count, s, r, t].imag = imag


cdef void unpack_int16(uint8_t *buf, np.complex128_t[:] csi_mem, int nfft,
                       bint flag):
    cdef int i, j
    if flag:
        for i in range(nfft):
            j = i * 4
            csi_mem[i].real = <double><int16_t>cu16l(buf[j+0], buf[j+1])
            csi_mem[i].imag = <double><int16_t>cu16l(buf[j+2], buf[j+3])
    else:
        for i in range(nfft):
            j = i * 4
            csi_mem[i].real = <double><int16_t>cu16b(buf[j+0], buf[j+1])
            csi_mem[i].imag = <double><int16_t>cu16b(buf[j+2], buf[j+3])


cdef void unpack_float(uint8_t *buf, np.complex128_t[:] csi_mem, int nfft,
                       int M, int E, int autoscale, bint flag):
    """N = M * R ^ E

    M: Mantissa
    R: Radix
    E: Exponent
    flag: little endian if ``True`` else big endian
    """
    cdef int i, s, e, shft, sgn
    cdef uint32_t h, m, b, x

    cdef int nbits = 10
    cdef int e_p = (1 << (E - 1))
    cdef int e_shift = 1
    cdef int e_zero = - M
    cdef int maxbit = -e_p
    cdef uint32_t k_tof_unpack_sgn_mask = (1 << 31)
    cdef uint32_t ri_mask = (1 << (M - 1)) - 1
    cdef uint32_t E_mask = (1 << E) - 1
    cdef uint32_t sgnr_mask = (1 << (E + 2 * M - 1))
    cdef uint32_t sgni_mask = sgnr_mask >> M
    cdef int8_t He[256]
    cdef int32_t Hout[512]
    cdef int32_t v_real, v_imag

    for i in range(nfft):
        if flag:
            h = cu32l(buf[4*i+0], buf[4*i+1], buf[4*i+2], buf[4*i+3])
        else:
            h = cu32b(buf[4*i+0], buf[4*i+1], buf[4*i+2], buf[4*i+3])

        v_real = <int32_t>((h >> (E + M)) & ri_mask)
        v_imag = <int32_t>((h >> E) & ri_mask)
        e = <int>(h & E_mask)
        if e >= e_p:
            e -= (e_p << 1)
        He[i] = <int8_t>e
        x = <uint32_t>v_real | <uint32_t>v_imag

        if autoscale and x:
            m = 0xffff0000
            b = 0xffff
            s = 16
            while s > 0:
                if x & m:
                    e += s
                    x >>= s
                s >>= 1
                m = (m >> s) & b
                b >>= s
            if e > maxbit:
                maxbit = e
        
        if h & sgnr_mask:
            v_real |= k_tof_unpack_sgn_mask
        if h & sgni_mask:
            v_imag |= k_tof_unpack_sgn_mask

        Hout[i<<1] = v_real
        Hout[(i<<1)+1] = v_imag

    shft = nbits - maxbit
    for i in range(nfft):
        e = He[(i*2 >> e_shift)] + shft
        sgn = 1
        v_real = Hout[i*2]
        if v_real & k_tof_unpack_sgn_mask:
            sgn = -1
            v_real &= ~k_tof_unpack_sgn_mask
        if e < e_zero:
            v_real = 0
        elif e < 0:
            e = -e
            v_real >>= e
        else:
            v_real <<= e
        v_real *= sgn

        e = He[(i*2+1 >> e_shift)] + shft
        sgn = 1
        v_imag = Hout[i*2+1]
        if v_imag & k_tof_unpack_sgn_mask:
            sgn = -1
            v_imag &= ~k_tof_unpack_sgn_mask
        if e < e_zero:
            v_imag = 0
        elif e < 0:
            e = -e
            v_imag >>= e
        else:
            v_imag <<= e
        v_imag *= sgn

        csi_mem[i].real = <double>v_real
        csi_mem[i].imag = <double>v_imag


cdef inline int8_t ccsi(uint8_t a, uint8_t b, uint8_t remainder):
    return ((a >> remainder) | (b << (8 - remainder))) & 0xff


cdef inline uint32_t cu32l(uint8_t a, uint8_t b, uint8_t c, uint8_t d):
    return a | (b << 8) | (c << 16) | (d << 24)


cdef inline uint32_t cu32b(uint8_t a, uint8_t b, uint8_t c, uint8_t d):
    return d | (c << 8) | (b << 16) | (a << 24)


cdef inline uint16_t cu16l(uint8_t a, uint8_t b):
    return a | (b << 8)


cdef inline uint16_t cu16b(uint8_t a, uint8_t b):
    return b | (a << 8)


cdef inline uint64_t cu64l(uint64_t a, uint64_t b, uint64_t c, uint64_t d,
                           uint64_t e, uint64_t f, uint64_t g, uint64_t h):
    return (a | (b << 8) | (c << 16) | (d << 24) |
            (e << 32) | (f << 40) | (g << 48) | (h << 56))


cdef inline uint64_t cu64b(uint64_t a, uint64_t b, uint64_t c, uint64_t d,
                           uint64_t e, uint64_t f, uint64_t g, uint64_t h):
    return (h | (g << 8) | (f << 16) | (e << 24) |
            (d << 32) | (c << 40) | (b << 48) | (a << 56))



"""csiread.Intel: implemented in pure Python"""

from timeit import default_timer
import os
import numpy as np


class Intel:
    def __init__(self, file, nrxnum=3, ntxnum=2, pl_len=0, if_report=True):
        self.file = file
        self.nrxnum = nrxnum
        self.ntxnum = ntxnum
        self.pl_len = pl_len    # useless
        self.if_report = if_report
        if not os.path.isfile(file):
            raise Exception("error: file does not exist, Stop!\n")

    def __getitem__(self, index):
        """Return contents of 0xbb packets"""
        ret = {
            "timestamp_low": self.timestamp_low[index],
            "bfee_count": self.bfee_count[index],
            "Nrx": self.Nrx[index],
            "Ntx": self.Ntx[index],
            "rssi_a": self.rssi_a[index],
            "rssi_b": self.rssi_b[index],
            "rssi_c": self.rssi_c[index],
            "noise": self.noise[index],
            "agc": self.agc[index],
            "perm": self.perm[index],
            "rate": self.rate[index],
            "csi": self.csi[index]
        }
        return ret

    def read(self):
        f = open(self.file, 'rb')
        if f is None:
            f.close()
            return -1

        lens = os.path.getsize(self.file)
        btype = np.int_
        self.timestamp_low = np.zeros([lens//95], dtype = btype)
        self.bfee_count = np.zeros([lens//95], dtype = btype)
        self.Nrx = np.zeros([lens//95], dtype = btype)
        self.Ntx = np.zeros([lens//95], dtype = btype)
        self.rssi_a = np.zeros([lens//95], dtype = btype)
        self.rssi_b = np.zeros([lens//95], dtype = btype)
        self.rssi_c = np.zeros([lens//95], dtype = btype)
        self.noise = np.zeros([lens//95], dtype = btype)
        self.agc = np.zeros([lens//95], dtype = btype)
        self.perm = np.zeros([lens//95, 3], dtype = btype)
        self.rate = np.zeros([lens//95], dtype = btype)
        self.csi = np.zeros([lens//95, 30, self.nrxnum, self.ntxnum], dtype = np.complex_)

        cur = 0
        count = 0
        while cur < (lens-3):
            temp = f.read(3)
            field_len = temp[1]+(temp[0]<<8)
            code = temp[2]
            cur += 3
            if code == 187:
                buf = f.read(field_len - 1)
                if len(buf) != field_len - 1:
                    break

                self.timestamp_low[count] = int.from_bytes(buf[:4], 'little')
                self.bfee_count[count] = int.from_bytes(buf[4:6], 'little')
                self.Nrx[count] = buf[8]
                self.Ntx[count] = buf[9]
                self.rssi_a[count] = buf[10]
                self.rssi_b[count] = buf[11]
                self.rssi_c[count] = buf[12]
                self.noise[count] = int.from_bytes(buf[13:14], 'little', signed=True)
                self.agc[count] = buf[14]
                self.rate[count] = int.from_bytes(buf[18:20], 'little')

                self.perm[count, 0] = buf[15] & 0x3
                self.perm[count, 1] = (buf[15] >> 2) & 0x3
                self.perm[count, 2] = (buf[15] >> 4) & 0x3

                index = 0
                payload = buf[20:]
                for i in range(30):
                    index += 3
                    remainder = index & 0x7
                    for j in range(buf[8]):
                        for k in range(buf[9]):
                            a = (payload[index // 8] >> remainder) | (payload[index // 8 + 1] << (8 - remainder)) & 0xff
                            b = (payload[index // 8 + 1] >> remainder) | (payload[index // 8 + 2] << (8 - remainder)) & 0xff
                            if a >= 128:
                                a -= 256
                            if b >= 128:
                                b -= 256
                            self.csi[count, i, self.perm[count, j], k] = a + b * 1.j
                            index += 16
                count += 1
            else:
                f.seek(field_len - 1, os.SEEK_CUR)
            cur += field_len - 1
        f.close()
        self.timestamp_low = self.timestamp_low[:count]
        self.bfee_count = self.bfee_count[:count]
        self.Nrx = self.Nrx[:count]
        self.Ntx = self.Ntx[:count]
        self.rssi_a = self.rssi_a[:count]
        self.rssi_b = self.rssi_b[:count]
        self.rssi_c = self.rssi_c[:count]
        self.noise = self.noise[:count]
        self.agc = self.agc[:count]
        self.perm = self.perm[:count, :]
        self.rate = self.rate[:count]
        self.csi = self.csi[:count, :, :, :]
        self.count = count

    def get_total_rss(self):
        """Calculates the Received Signal Strength (RSS) in dBm from CSI"""
        rssi_mag = np.zeros_like(self.rssi_a, dtype=np.float)
        rssi_mag += self.__dbinvs(self.rssi_a)
        rssi_mag += self.__dbinvs(self.rssi_b)
        rssi_mag += self.__dbinvs(self.rssi_c)
        ret = self.__db(rssi_mag) - 44 - self.agc
        return ret

    def get_scaled_csi(self):
        """Converts CSI to channel matrix H"""
        csi = self.csi
        csi_sq = (csi * csi.conj()).real
        csi_pwr = np.sum(csi_sq, axis=(1, 2, 3))
        rssi_pwr = self.__dbinv(self.get_total_rss())

        scale = rssi_pwr / (csi_pwr / 30)

        noise_db = self.noise
        thermal_noise_pwr = self.__dbinv(noise_db)
        thermal_noise_pwr[noise_db == -127] = self.__dbinv(-92)

        quant_error_pwr = scale * (self.Nrx * self.Ntx)
        total_noise_pwr = thermal_noise_pwr + quant_error_pwr

        ret = self.csi * np.sqrt(scale / total_noise_pwr).reshape(-1, 1, 1, 1)
        ret[self.Ntx == 2] *= np.sqrt(2)
        ret[self.Ntx == 3] *= np.sqrt(self.__dbinv(4.5))
        return ret

    def get_scaled_csi_sm(self):
        """Converts CSI to channel matrix H(Not test this method)

        This version undoes Intel's spatial mapping to return the pure
        MIMO channel matrix H.

        """
        ret = self.get_scaled_csi()
        ret = self.__remove_sm(ret)
        return ret

    def __dbinvs(self, x):
        """Convert from decibels specially"""
        ret = np.power(10, x / 10)
        ret[ret == 1] = 0
        return ret

    def __dbinv(self, x):
        """Convert from decibels"""
        ret = np.power(10, x / 10)
        return ret

    def __db(self, x):
        """Calculates decibels"""
        ret = 10 * np.log10(x)
        return ret

    def __remove_sm(self, scaled_csi):
        """Actually undo the input spatial mapping"""
        sm_1 = 1
        sm_2_20 = np.array([[1, 1],
                            [1, -1]]) / np.sqrt(2)
        sm_2_40 = np.array([[1, 1j],
                            [1j, 1]]) / np.sqrt(2)
        sm_3_20 = np.array([[-2 * np.pi / 16, -2 * np.pi / (80 / 33), 2 * np.pi / (80 / 3)],
                            [ 2 * np.pi / (80 / 23), 2 * np.pi / (48 / 13), 2 * np.pi / (240 / 13)],
                            [-2 * np.pi / (80 / 13), 2 * np.pi / (240 / 37), 2 * np.pi / (48 / 13)]])
        sm_3_20 = np.exp(1j * sm_3_20) / np.sqrt(3)
        sm_3_40 = np.array([[-2 * np.pi / 16, -2 * np.pi / (80 / 13), 2 * np.pi / (80 / 23)],
                            [-2 * np.pi / (80 / 37), -2 * np.pi / (48 / 11), -2 * np.pi / (240 / 107)],
                            [ 2 * np.pi / (80 / 7), -2 * np.pi / (240 / 83), -2 * np.pi / (48 / 11)]])
        sm_3_40 = np.exp(1j * sm_3_40) / np.sqrt(3)
    
        ret = scaled_csi
        for i in range(self.count):
            M = self.Ntx[i]
            if (int(self.rate[i]) & 2048) == 2048:
                if M == 3:
                    sm = sm_3_40
                elif M == 2:
                    sm = sm_2_40
                else:
                    sm = sm_1
            else:
                if M == 3:
                    sm = sm_3_20
                elif M == 2:
                    sm = sm_2_20
                else:
                    sm = sm_1
            ret[i, :, :, :M] = ret[i, :, :, :M].dot(sm.T.conj())
        return ret


CSI = Intel


if __name__ == '__main__':
    last = default_timer()
    csifile = '../material/5300/dataset/sample_0x5_64_3000.dat'
    csidata = Intel(csifile, nrxnum=3, ntxnum=1, pl_len=0, if_report=True)
    csidata.read()
    print(default_timer() - last, 's')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""CSI server: simulation of real-time packet sending

Usage:
    Intel5300: python3 csiserver.py ../material/5300/dataset/sample_0x5_64_3000.dat 3000 1000
    Atheros: python3 csiserver.py ../material/atheros/dataset/ath_csi_1.dat 100 100000
    Nexmon: sudo python3 csiserver_nexmon.py ../material/nexmon/dataset/example.pcap 12 10000
"""

import argparse
import os
import socket
import time
from utils import infer_device


def intel_server(csifile, number, delay):
    """intel server

    Args:
        csifile: csi smaple file
        number: packets number, unlimited if number=0
        delay: packets rate(us), the sending rate is inaccurate because of `sleep`

    Note:
        set address for remoting connection
    """
    # config
    address_src = ('127.0.0.1', 10086)
    address_des = ('127.0.0.1', 10010)

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.bind(address_src)

    f = open(csifile, 'rb')
    lens = f.seek(0, os.SEEK_END)
    f.seek(0, os.SEEK_SET)

    cur = 0
    count = 0

    print("sending")
    while True:
        if cur >= (lens - 3):
            f.seek(0, os.SEEK_SET)
            cur = 0

        # data
        field_len = int.from_bytes(f.read(2), byteorder='big')
        code = int.from_bytes(f.read(1), byteorder='little')
        f.seek(-1, os.SEEK_CUR)
        data = bytearray(f.read(field_len))

        # set timestamp_low
        if code == 0xbb:
            time.sleep(delay/1000000)
            timestamp_low = int(time.time() * 1000000) & 0xFFFFFFFF
            data[1:5] = timestamp_low.to_bytes(4, 'little')

        s.sendto(data, address_des)

        cur += (field_len + 2)
        if code == 0xbb:
            count += 1
            if count % 1000 == 0:
                print(".", end="", flush=True)
            if count % 50000 == 0:
                print(count//1000, 'K', flush=True)
            if number != 0 and count >= number:
                break

    s.close()
    f.close()
    print()


def atheros_server(csifile, number, delay, endian):
    """atheros server

    Args:
        csifile: csi smaple file
        number: packets number, unlimited if number=0
        delay: packets rate(us), the sending rate is inaccurate because of `sleep`
        endian: the endian of csifile.

    Note:
        set address for remoting connection
    """
    # config
    address_src = ('127.0.0.1', 10086)
    address_des = ('127.0.0.1', 10010)

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.bind(address_src)

    f = open(csifile, 'rb')
    lens = f.seek(0, os.SEEK_END)
    f.seek(0, os.SEEK_SET)

    cur = 0
    count = 0

    print("sending")
    while True:
        if cur >= (lens - 4):
            f.seek(0, os.SEEK_SET)
            cur = 0

        # data
        field_len = int.from_bytes(f.read(2), byteorder=endian)
        data = bytearray(f.read(field_len))

        # set timestamp_low
        time.sleep(delay/1000000)
        timestamp_low = int(time.time() * 1000000) & 0xFFFFFFFFFFFFFFFF
        data[:8] = timestamp_low.to_bytes(8, endian)

        s.sendto(data, address_des)

        cur += (field_len + 2)

        count += 1
        if count % 1000 == 0:
            print(".", end="", flush=True)
        if count % 50000 == 0:
            print(count//1000, 'K', flush=True)
        if number != 0 and count >= number:
            break

    s.close()
    f.close()
    print()


def nexmon_server(csifile, number, delay):
    s = socket.socket(socket.PF_PACKET, socket.SOCK_RAW, socket.IPPROTO_IP)
    s.bind(("wlp4s0", socket.IPPROTO_IP))

    f = open(csifile, 'rb')
    magic = f.read(4)
    f.seek(20, os.SEEK_CUR)
    lens = os.path.getsize(csifile)
    endian = 'big' if magic in [b"\xa1\xb2\xc3\xd4", b"\xa1\xb2\x3c\x4d"] else "little"
    cur = 24
    count = 0

    while True:
        if cur >= (lens - 4):
            f.seek(24, os.SEEK_SET)
            cur = 0

        caplen = int.from_bytes(f.read(16)[8:12], byteorder=endian)
        if f.read(42)[6:12] == b'NEXMON':
            time.sleep(delay/1000000)
            f.seek(-42, os.SEEK_CUR)
            data = f.read(caplen)
            s.send(data)

            count += 1
            if count % 1000 == 0:
                print(".", end="", flush=True)
            if count % 50000 == 0:
                print(count//1000, 'K', flush=True)
            if number != 0 and count >= number:
                break
        else:
            f.seek(caplen - 42, os.SEEK_CUR)
        cur += (caplen + 16)

    f.close()
    s.close()
    print()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('csifile', type=str, help='csi smaple file')
    parser.add_argument('number', type=int, help='packets number')
    parser.add_argument('delay', type=int, help='delay in us')
    p = parser.parse_args()

    device = infer_device(p.csifile)

    if device == "Intel":
        intel_server(p.csifile, p.number, p.delay)
    elif device == "Atheros":
        atheros_server(p.csifile, p.number, p.delay, 'little')
    else:
        nexmon_server(p.csifile, p.number, p.delay)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Plot CSI(nexmon_csi) in real time

Usage:
    1. sudo python3 csirealtime_nexmon.py
    2. sudo python3 csiserver.py ../material/5300/dataset/example.pcap 4 10000

Important:
    1. Nexmon.pmsg() is experimental, it may be modified in the future.
    The first arguement is raw packet, however, I'd like it to be udp packet. 
    2. csirealtime_nexmon.py and csiserver.py(Nexmon) need root or sudo 
    permissions. Only work on Linux.
"""

import socket
import threading

import csiread
import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np

subcarrier_num = 256
cache_data1 = [np.nan] * subcarrier_num
mutex = threading.Lock()


class GetDataThread(threading.Thread):
    def run(self):
        update_background()


def update_background():
    csidata = csiread.Nexmon(None, chip='4358', bw=80)

    # config
    global cache_data1, mutex
    count = 0

    with socket.socket(socket.PF_PACKET, socket.SOCK_RAW, socket.htons(0x3)) as s:
        while True:
            data, address_src = s.recvfrom(4096)
            msg_len = len(data)

            code = csidata.pmsg(data)
            if code == 0xf100:
                mutex.acquire()
                cache_data1 = np.fft.ifftshift(csidata.csi[0])
                mutex.release()
                count += 1
                if count % 100 == 0:
                    print('receive %d bytes [msgcnt=%u]' % (msg_len, count))


def realtime_plot():
    fig, ax = plt.subplots()
    plt.title('csi-amplitude')
    plt.xlabel('subcarrier')
    plt.ylabel('amplitude')
    ax.set_ylim(0, 4000)
    ax.set_xlim(0, subcarrier_num)
    x = np.arange(0, subcarrier_num)

    line1,  = ax.plot(x, np.abs(cache_data1), linewidth=1.0, label='subcarrier_256')
    plt.legend()

    def init():
        line1.set_ydata([np.nan] * subcarrier_num)
        return line1,

    def animate(i):
        global cache_data1, mutex
        mutex.acquire()
        line1.set_ydata(np.abs(cache_data1))
        mutex.release()
        return line1,

    ani = animation.FuncAnimation(fig, animate, init_func=init, interval=1000/25, blit=True)
    plt.show()


if __name__ == '__main__':
    task = GetDataThread()
    task.start()
    realtime_plot()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""CSI ratioA: Plot CSI ratio of two antennas in real time(Linux 802.11n CSI Tool)

Usage:
    1. python3 csiratioA.py
    2. python3 csiserver.py ../material/5300/dataset/sample_0x5_64_3000.dat 3000 1000

Note:
    1. This example is much faster than csiratioB.py, and less packets loss.
    2. You need right data to get beautiful result

Ref:
    csi ratio: [FarSense: Pushing the Range Limit of WiFi-based Respiration Sensing
        with CSI Ratio of Two Antennas](https://arxiv.org/pdf/1907.03994v1.pdf)
"""

import os
import socket
import sys

import csiread
import numpy as np
import pyqtgraph as pg
from PyQt5.QtCore import QMutex, Qt, QThread, QTimer, pyqtSlot
from PyQt5.QtGui import QApplication
from PyQt5.QtWidgets import QVBoxLayout, QWidget

os.environ['QT_SCALE_FACTOR'] = '1'

cache = np.zeros([30, 800])
mutex = QMutex()


class GetDataThread(QThread):
    def __init__(self, parent):
        super(GetDataThread, self).__init__(parent)

    def stop(self):
        self.requestInterruption()
        self.exit()

    def run(self):
        """get data in real time

        Note:
            If you want to run this script on the host with Intel 5300 NIC, rewrite code as
            csiuserspace.py
        """
        csidata = csiread.Intel(None, 3, 2)

        # config
        global cache, mutex
        count = 0
        address_src = ('127.0.0.1', 10086)
        address_des = ('127.0.0.1', 10010)

        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.bind(address_des)
            s.settimeout(30/1000)
            while not self.isInterruptionRequested():
                try:
                    data, address_src = s.recvfrom(4096)
                except socket.timeout:
                    continue
                msg_len = len(data)

                code = csidata.pmsg(data)
                if code == 0xbb:
                    csi = csidata.get_scaled_csi_sm()
                    scaled_csi_sm = np.abs(csi[0, :, 0, 0] / csi[0:, :, 1, 0])
                    scaled_csi_sm[scaled_csi_sm==np.inf] = 0
                    mutex.lock()
                    cache[:, :-1] = cache[:, 1:]
                    cache[:, -1] = scaled_csi_sm
                    mutex.unlock()
                    count += 1
                    if count % 100 == 0:
                        print('receive %d bytes [msgcnt=%u], seq=%d' % (msg_len, count, csidata.seq))


class MainWindow(QWidget):
    def __init__(self, parent=None, flags=Qt.WindowFlags()):
        super(MainWindow, self).__init__(parent=parent, flags=flags)
        self.configUI()
        self.configLayout()
        self.show()

    def configUI(self):
        screen_size = QApplication.primaryScreen().size()
        self.setWindowTitle("CSIRatioA")
        self.setWindowFlags(Qt.WindowCloseButtonHint | Qt.WindowMinimizeButtonHint | Qt.WindowMaximizeButtonHint)
        self.resize(800, 450)
        self.move(screen_size.width() - 800 - 100, 100)

        # pygtgrapg global config
        pg.setConfigOptions(antialias=True)                 # Antialiasing
        pg.setConfigOptions(foreground=(232, 232, 232))     # Black
        pg.setConfigOptions(background=(25, 25, 25))        # White

        # plot area
        self.win = pg.GraphicsWindow()

        # plot init
        self.initPlot()

        # update cache
        self.task = GetDataThread(self)
        self.task.start()

        # plot refresh
        self.PlotTimer()

    def configLayout(self):
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.win)

    def initPlot(self):
        lens = cache.shape[-1]

        p = self.win.addPlot(title="Ratio of CSI(30x)")
        p.showGrid(x=False, y=True)
        p.setLabel('left', "Amplitude", units='A')
        p.setLabel('bottom', "Packets", units='')
        p.enableAutoRange('xy', False)
        p.setXRange(-int(lens/2), int(lens/2), padding=0.01)
        p.setYRange(-10, 70, padding=0.01)
        p.hideAxis('left')
        # p.addLegend()

        # curves
        self.curves = []
        for i in range(len(cache)):
            color = (int(np.sin(i * (2 * np.pi) / 30 + 2) * 127 + 128),
                     int(np.cos(i * (2 * np.pi) / 30 - 1) * 127 + 128),
                     int(np.sin(i * (2 * np.pi) / 30 - 2) * 127 + 128),
                     127)
            self.curves.append(p.plot(pen=color, name='subcarrier=%d' % (i)))

        # x axis
        self.X = np.linspace(-int(lens/2), int(lens/2), lens)

    def PlotTimer(self):
        self.timer = QTimer()
        self.timer.setTimerType(0)
        self.timer.timeout.connect(self.updatePlot)
        self.timer.start(1000//30)

    @pyqtSlot()
    def updatePlot(self):
        global cache, mutex
        mutex.lock()
        for i in range(len(cache)):
            xscale = 1 - i / 200.
            self.curves[i].setData(xscale * self.X, i/2 + cache[i])
        mutex.unlock()

    def closeEvent(self, event):
        self.task.stop()
        event.accept()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = MainWindow()
    sys.exit(app.exec_())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Drcsi: `Decimeter Ranging with Channel State Information`

Usage:
    python3 csiDrcsi.py

Important:
    1. alg_3: it's not implemented. I can't understand *Algorithm 3 CFO + Weak Stream Removal*.
    2. alg_2: Is N_sto right?

Note:
    no dataset available

Ref:
    Drcsi: [Decimeter Ranging with Channel State Information](https://arxiv.org/pdf/1902.09652.pdf)
"""

import csiread
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from scipy.fftpack import fft, ifft, fftshift, ifftshift
from scipy.signal import find_peaks
from scipy.stats import mode
from utils import scidx, calib, phy_ifft, phy_fft


def fig_2(csidata, index=34):
    """fig_2

    Note:
        Fig. 2a uses data collected from a setup where transmitter and receiver arrays
        directly face each other whereas, in Fig. 2b, the receiver is rotated by 90 degrees

    Ref:
        [Power Delay Profile](https://www.gaussianwaves.com/2014/07/power-delay-profile/)
        [From Channel Impulse Response (CIR) to Power Delay Profile (PDP)](https://www.mathworks.com/matlabcentral/answers/44530-from-channel-impulse-response-cir-to-power-delay-profile-pdp)
    """
    csi = csidata.csi

    pdp = np.power(np.abs(phy_ifft(csi, axis=1)), 2.0) / csi.shape[1]
    norm = lambda x: x / np.max(x)

    plt.figure()
    plt.plot(norm(pdp[index, :12].reshape(12, -1)))
    plt.title('Anechoic Chamber: $N_{SS}=%d$' % (sum(csi[index, 0, 0] != 0)))
    plt.xlabel('sample index')
    plt.ylabel('PDP')
    plt.tight_layout()
    plt.show()


def fig_4(athdata, index=26):
    """fig_4

    Ref:
        [FFT Length (802.11n/ac/ax)](http://rfmw.em.keysight.com/wireless/helpfiles/89600B/WebHelp/Subsystems/wlan-mimo/Content/mimo_adv_fftsze.htm)
    """
    csi = athdata.csi

    # Fig.4(a)
    phase = np.unwrap(np.angle(csi), axis=1)
    phase = calib(phase, 20, 1)
    plt.subplot(3, 1, 1)
    plt.plot(phase[index].reshape(csi.shape[1], -1))

    # Fig.4(b)
    amplidute = 10 * np.log10(np.abs(csi) + 1 / 10 ** 6)    # db
    plt.subplot(3, 1, 2)
    plt.plot(amplidute[index].reshape(csi.shape[1], -1))

    # Fig.4(c)
    # scipy.fftpack.fft or phy_fft ?
    A = np.log10(np.abs(fft(csi, n=64, axis=1)))
    plt.subplot(3, 1, 3)
    plt.plot(A[index].reshape(64, -1), '*--')

    plt.tight_layout()
    plt.show()


def fig_6(csidata):
    """fig_6"""
    csi = csidata.csi
    s_index = scidx(20, 2)

    phase = np.unwrap(np.angle(csi[:1000, :, 0, 0])).T
    phase_diff = np.diff(phase)
    phase_diff -= phase_diff.mean(axis=0)
    phase_mean_0 = phase_diff.mean(axis=-1)

    plt.figure()
    plt.subplot(2, 1, 1)
    plt.plot(s_index, phase_diff, 'y+', linewidth=0.3)
    plt.plot(s_index, phase_mean_0, 'r-', linewidth=2.0)
    patch_1 = mpatches.Patch(color='yellow', label=':1000_r0t0')
    patch_2 = mpatches.Patch(color='red', label=':1000_r0t0_phase_mean_0')
    plt.legend(handles=[patch_1, patch_2])
    plt.title('fig6b: csi-phase_diff')
    plt.xlabel('subcarriers')
    plt.ylabel('phase_diff')

    plt.subplot(2, 1, 2)
    plt.hist(phase_diff[12], bins=250, histtype='stepfilled', density=False, color='red', alpha=0.5)
    plt.hist(phase_diff[13], bins=250, histtype='stepfilled', density=False, color='green', alpha=0.5)
    patch_1 = mpatches.Patch(color='red', label=':1000_s12r0t0')
    patch_2 = mpatches.Patch(color='green', label=':1000_s13r0t0')
    plt.legend(handles=[patch_1, patch_2])
    plt.title('fig6c: csi-phase_diff')
    plt.xlabel('phase_diff')
    plt.ylabel('packets count')
    plt.tight_layout()
    plt.show()


def formula_14(csi, Nrx=0, Ntx=0):
    """formula_14"""
    return - np.angle(np.mean(csi[:, :-1, Nrx, Ntx] * csi[:, 1:, Nrx, Ntx].conj(), axis=1))


def formula_15(csi):
    """formula_15"""
    return - np.angle(np.mean(csi[:, :-1] * csi[:, 1:].conj(), axis=(1, 2, 3)))


def formula_16(csi):
    """formula_16"""
    k = scidx(20, 2)[:, np.newaxis, np.newaxis]
    phi = formula_15(csi)[:, np.newaxis, np.newaxis, np.newaxis]
    return np.exp(1j * phi * k) * csi


def derotate_linear(csi, bw=20, ng=2):
    """Windowing/CP removal effect compensation"""
    k = scidx(bw, ng)[:, np.newaxis, np.newaxis]
    phi = -7e-5 * k ** 3 + 3e-5 * k ** 2 + 0.05 * k
    return np.exp(-1j * phi) * csi


def derotate_formual16(csi, phi, bw=20, ng=2):
    """De-rotate CSI phase using (16)"""
    k = scidx(bw, ng)[:, np.newaxis, np.newaxis]
    phi = phi[:, np.newaxis, np.newaxis, np.newaxis]
    return np.exp(1j * phi * k) * csi


def alg_1(csi):
    """alg_1"""
    csi = derotate_linear(csi)
    phi = formula_15(csi)
    csi = derotate_formual16(csi, phi)
    return csi


def alg_2(csi):
    """alg_2

    Note: N_p = 1
    """
    p = np.power(np.abs(phy_ifft(csi, axis=1)), 2.0)
    # find_peak() is better
    s = p[:, :20].argmax(axis=1).reshape(csi.shape[0], -1)
    N_sto = scidx(20, 2)[mode(s, axis=1)[0][:, 0]]
    N_sc = 64
    csi = derotate_formual16(csi, -2 * np.pi * N_sto / N_sc)
    return csi


def alg_3(csi, index):
    """alg_3"""
    pass
    # P = 1
    # Np = 10
    # for i in range(Np):
    #     P *= csi[index+i]
    # ret = np.power(P, 1 / Np)

    # R = np.linalg.norm(ret)
    # for jrx in range(csi.shape[2]):
    #     for jss in range(csi.shape[3]):
    #         for k in range(csi.shape[1]):
    #             if np.linalg.norm(ret[k, jrx, jss]) / R < 0.1:
    #                 ret[k, jrx, jss] = np.nan

    # return ret


def algshow(csidata):
    s_index = scidx(20, 2)
    csi = csidata.csi[:500]

    plt.figure()
    
    # alg_1
    csi = alg_1(csi)
    phase = np.unwrap(np.angle(csi), axis=1)
    phase = calib(phase)
    phase0 = phase
    plt.subplot(3, 1, 1)
    plt.plot(s_index, phase[200:300, :, 0, 0].T)

    # alg_2
    csi = alg_2(csi)
    phase = np.unwrap(np.angle(csi), axis=1)
    phase = calib(phase)
    plt.subplot(3, 1, 2)
    plt.plot(s_index, phase[200:300, :, 0, 0].T)

    # alg_3
    # csicopy = csi.copy()
    # for i in range(100, 200):
    #     csicopy[i] = alg_3(csi, i)
    # phase = np.unwrap(np.angle(csicopy), axis=1)
    # phase = calib(phase)
    # plt.subplot(3, 1, 3)
    # plt.plot(s_index, phase[100:200, :, 0, 0].T)
    plt.show()


if __name__ == "__main__":
    csidata = csiread.Intel("../material/5300/dataset/sample_0x5_64_3000.dat", ntxnum=1, if_report=False)
    athdata = csiread.Atheros("../material/atheros/dataset/ath_csi_1.dat", if_report=False)
    csidata.read();csidata.get_scaled_csi(True)
    athdata.read()

    fig_2(csidata)
    fig_4(athdata)
    fig_6(csidata)
    # algshow(csidata)



"""csiread.Atheros: implemented in pure Python"""

import numpy as np
import os
from timeit import default_timer


class Atheros:
    def __init__(self, file, nrxnum=3, ntxnum=2, tones=56, pl_len=0, if_report=True):
        """Parameter initialization."""
        self.file = file
        self.nrxnum = nrxnum
        self.ntxnum = ntxnum
        self.tones = tones
        self.pl_len = pl_len
        self.if_report = if_report

        if not os.path.isfile(file):
            raise Exception("error: file does not exist, Stop!\n")

    def read(self, endian='little'):
        f = open(self.file, 'rb')
        if f is None:
            f.close()
            return -1

        lens = os.path.getsize(self.file)
        btype = np.int_
        self.timestamp = np.zeros([lens//420])
        self.csi_len = np.zeros([lens//420], dtype=btype)
        self.tx_channel = np.zeros([lens//420], dtype=btype)
        self.err_info = np.zeros([lens//420], dtype=btype)
        self.noise_floor = np.zeros([lens//420], dtype=btype)
        self.Rate = np.zeros([lens//420], dtype=btype)
        self.bandWidth = np.zeros([lens//420], dtype=btype)
        self.num_tones = np.zeros([lens//420], dtype=btype)
        self.nr = np.zeros([lens//420], dtype=btype)
        self.nc = np.zeros([lens//420], dtype=btype)
        self.rssi = np.zeros([lens//420], dtype=btype)
        self.rssi_1 = np.zeros([lens//420], dtype=btype)
        self.rssi_2 = np.zeros([lens//420], dtype=btype)
        self.rssi_3 = np.zeros([lens//420], dtype=btype)
        self.payload_len = np.zeros([lens//420], dtype=btype)
        self.csi = np.zeros([lens//420, self.tones, self.nrxnum, self.ntxnum], dtype=np.complex128)
        self.payload = np.zeros([lens//420, self.pl_len], dtype=btype)

        cur = 0
        count = 0
        while cur < (lens - 4):
            field_len = int.from_bytes(f.read(2), byteorder=endian)
            cur += 2
            if (cur + field_len) > lens:
                break

            self.timestamp[count] = int.from_bytes(f.read(8), byteorder=endian)
            cur += 8
            self.csi_len[count] = int.from_bytes(f.read(2), byteorder=endian)
            cur += 2
            self.tx_channel[count] = int.from_bytes(f.read(2), byteorder=endian)
            cur += 2
            self.err_info[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.noise_floor[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.Rate[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.bandWidth[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.num_tones[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.nr[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.nc[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.rssi[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.rssi_1[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.rssi_2[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.rssi_3[count] = int.from_bytes(f.read(1), byteorder=endian)
            cur += 1
            self.payload_len[count] = int.from_bytes(f.read(2), byteorder=endian)
            cur += 2

            c_len = self.csi_len[count]
            if c_len > 0:
                csi_buf = f.read(c_len)
                self.csi[count] = self.__read_csi(csi_buf, self.nr[count], self.nc[count], self.num_tones[count])
                cur += c_len
            else:
                self.csi[count] = None

            pl_len = self.payload_len[count]
            pl_stop = min(pl_len, self.pl_len, 0)
            if pl_len > 0:
                self.payload[count, :pl_stop] = bytearray(f.read(pl_len))[:pl_stop]
                cur += pl_len
            else:
                self.payload[count, :pl_stop] = 0

            if (cur + 420 > lens):
                count -= 1
                break
            count += 1

        self.timestamp = self.timestamp[:count]
        self.csi_len = self.csi_len[:count]
        self.tx_channel = self.tx_channel[:count]
        self.err_info = self.err_info[:count]
        self.noise_floor = self.noise_floor[:count]
        self.Rate = self.Rate[:count]
        self.bandWidth = self.bandWidth[:count]
        self.num_tones = self.num_tones[:count]
        self.nr = self.nr[:count]
        self.nc = self.nc[:count]
        self.rssi = self.rssi[:count]
        self.rssi_1 = self.rssi_1[:count]
        self.rssi_2 = self.rssi_2[:count]
        self.rssi_3 = self.rssi_3[:count]
        self.payload_len = self.payload_len[:count]
        self.csi = self.csi[:count]
        self.payload = self.payload[:count]

        f.close()

    def __read_csi(self, csi_buf, nr, nc, num_tones):
        csi = np.zeros([self.tones, self.nrxnum, self.ntxnum], dtype=np.complex128)

        bits_left = 16
        bitmask = (1 << 10) - 1

        idx = 0
        h_data = csi_buf[idx]
        idx += 1
        h_data += (csi_buf[idx] << 8)
        idx += 1
        curren_data = h_data & ((1 << 16) - 1)

        for k in range(num_tones):
            for nc_idx in range(nc):
                for nr_idx in range(nr):
                    # imag
                    if (bits_left - 10) < 0:
                        h_data = csi_buf[idx]
                        idx += 1
                        h_data += (csi_buf[idx] << 8)
                        idx += 1
                        curren_data += h_data << bits_left
                        bits_left += 16
                    imag = curren_data & bitmask
                    imag = self.__signbit_convert(imag, 10)

                    bits_left -= 10
                    curren_data = curren_data >> 10
                    # real
                    if (bits_left - 10) < 0:
                        h_data = csi_buf[idx]
                        idx += 1
                        h_data += (csi_buf[idx] << 8)
                        idx += 1
                        curren_data += h_data << bits_left
                        bits_left += 16
                    real = curren_data & bitmask
                    real = self.__signbit_convert(real, 10)

                    bits_left -= 10
                    curren_data = curren_data >> 10
                    # csi
                    csi[k, nr_idx, nc_idx] = real + imag * 1j

        return csi

    def __signbit_convert(self, data, maxbit):
        if data & (1 << (maxbit - 1)):
            data -= (1 << maxbit)
        return data


if __name__ == '__main__':
    last = default_timer()
    csifile = '../material/atheros/dataset/ath_csi_1.dat'
    csidata = Atheros(csifile, nrxnum=3, ntxnum=2, pl_len=10, if_report=True)
    csidata.read()
    print(default_timer() - last, 's')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""CSI MemoryError: Solution to the MemoryError

csiread may raise MemoryError when reading a super large file. It can be solved
by setting 'bufsize'. However, this error rarely happens.

Note:
    csiread.Intel needs to know the count of packets in the csifile before
    parsing. It estimates packets count by the following method.

    ----------------------------------------------------------------------
    | packet type | field_len | code | header |       payload       | CRC |
    |     0xbb    |     2     |   1  |   20   | 12 + 60 * Ntx * Nrx |     |
    |     0xc1    |     2     |   1  | 4 + 20 |    pl_size - 28     |  4  |
    ----------------------------------------------------------------------
    csifile_size = os.path.getsize(csifile)
    packet_number = csifile_size // packet_size

    packet_size is set to a constant value of 95. it is not a good choice.
    There are two known limitations here: 1. packet size >= 95 is necessary,
    otherwise it will throw `IndexError: Out of bounds on buffer access
    (axis 0)`. 2. if packet size > 95, the more packets, the more excess memory
    allocated. Excess memory is so large that MemoryError is throwed. 
    
    csiread.Atheros has the same issue. (Atheros.packet_size = 420). csiread.Nexmon
    does not have this issue, but it calculates the count of packets by the
    pre-read. It takes extra time.

    12 + 60 * Ntx * Nrx = (30 * (Nrx * Ntx * 8 * 2 + 3) + 7) / 8 :
    ===========================================================================
    Grouping is a method that reduces the size of the CSI Report field by
    reporting a single value for each group of Ng adjacent subcarriers. With
    grouping, the size of the CSI Report field is Nr×8+Ns×(3+2×Nb×Nc×Nr) bits,
    where the number of subcarriers sent, Ns, is a function of Ng and whether
    matrices for 40 MHz or 20 MHz are sent. The value of Ns and the specific
    carriers for which matrices are sent are shown in Table 7-25f. If the size
    of the **CSI Report field** is not an integral multiple of 8 bits, up to 7
    zeros are appended to the end of the report to make its size an integral
    multiple of 8 bits.
    ===========================================================================
ref:
    [IEEE 802.11n Standard](Link has expired)
"""

from functools import wraps

import csiread


def info(func):
    """csidata info"""
    @wraps(func)
    def wrapper(*arg, **argv):
        csidata = func(*arg, **argv)
        s = "{}: csidata.Nrx.size={} csidata.Nrx.base.size={}"
        s = s.format(func.__name__, csidata.Nrx.size, csidata.Nrx.base.size)
        print(s)
        return csidata
    return wrapper


@info
def read_bf_fileA(csifile, pk_num):
    csidata = csiread.Intel(csifile, nrxnum=3, ntxnum=2, pl_size=0, if_report=False, bufsize=pk_num)
    csidata.read()
    return csidata


@info
def read_bf_fileB(csifile):
    csidata = csiread.Intel(csifile, nrxnum=3, ntxnum=2, pl_size=0, if_report=False)
    csidata.read()
    return csidata


if __name__ == "__main__":
    csifile = "../material/5300/dataset/sample_0x1_ap.dat"
    csidataA = read_bf_fileA(csifile, pk_num=600)
    csidataB = read_bf_fileB(csifile)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""csitable: view the contents of packets using beautiful table

Note: run in jupyter notebook to get sample1.png.
"""

import csiread
import pandas as pd

csifile = "../material/5300/dataset/sample_0x5_64_3000.dat"
csidata = csiread.Intel(csifile, ntxnum=1, pl_size=6, if_report=False)
csidata.read()

# pd.set_option('display.max_rows', None)
# pd.set_option('display.max_columns', None)

stringify = lambda data, space: [space.join([hex(per)[2:].zfill(2).upper() 
                                             for per in data[index]])
                                 for index in range(len(data))]

data = csidata[:]
del data['csi']
del data['perm']
data['perm'] = csidata.perm.tolist()
data['csi.shape'] = [csidata.csi.shape[1:]]*csidata.count
data['fc'] = csidata.fc
data['dur'] = csidata.dur
data['seq'] = csidata.seq
data['addr_src'] = stringify(csidata.addr_src, ":")
data['addr_des'] = stringify(csidata.addr_des, ":")
data['addr_bssid'] = stringify(csidata.addr_bssid, ":")
data['payload[:6]'] = stringify(csidata.payload[:, :6], " ")

csitable = pd.DataFrame(data)
print(csitable[:10])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""CSI ratioB: Plot CSI ratio of two antennas in real time(Linux 802.11n CSI Tool)

Usage:
    1. python3 csiratioB.py
    2. python3 csiserver.py ../material/5300/dataset/sample_0x5_64_3000.dat 3000 10000

Note:
    1. this is just a demo, because it's too slow :(
    2. You need right data to get beautiful result
    3. csiserver.py: Use a large delay value to avoid too much packets loss

Ref:
    realtime_plot: [MATPLOTLIB UNCHAINED](https://matplotlib.org/3.1.3/gallery/animation/unchained.html#matplotlib-unchained)
    csi ratio: [FarSense: Pushing the Range Limit of WiFi-based Respiration Sensing with CSI Ratio of Two Antennas](https://arxiv.org/pdf/1907.03994v1.pdf)
"""

import socket
import threading

import csiread
import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np

plt.rcParams['toolbar'] = "None"

cache = np.zeros([30, 800])
mutex = threading.Lock()


class GetDataThread(threading.Thread):
    def __init__(self):
        super(GetDataThread, self).__init__()
        self.__state = True

    def run(self):
        """get data in real time

        Note:
            If you want to run this script on the host with Intel 5300 NIC, rewrite code as
            csiuserspace.py
        """
        csidata = csiread.Intel(None, 3, 2)

        # config
        global cache, mutex
        count = 0
        address_src = ('127.0.0.1', 10086)
        address_des = ('127.0.0.1', 10010)

        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.bind(address_des)
            s.settimeout(0.1)
            while self.__state:
                try:
                    data, address_src = s.recvfrom(4096)
                except socket.timeout:
                    continue
                msg_len = len(data)

                code = csidata.pmsg(data)
                if code == 0xbb:
                    csi = csidata.get_scaled_csi_sm()
                    scaled_csi_sm = np.abs(csi[0, :, 0, 0] / csi[0:, :, 1, 0])
                    mutex.acquire()
                    cache[:, :-1] = cache[:, 1:]
                    cache[:, -1] = scaled_csi_sm
                    mutex.release()
                    count += 1
                    if count % 100 == 0:
                        print('receive %d bytes [msgcnt=%u], seq=%d' % (msg_len, count, csidata.seq))

    def stop(self):
        self.__state = False


def realtime_plot():
    fig = plt.figure(figsize=(16/2, 9/2), facecolor='black')
    ax = plt.subplot(111, frameon=False)
    ax.set_ylim(-10, 70)
    # ax.set_xticks([])
    ax.set_yticks([])
    ax.tick_params(axis='x', colors='white')
    ax.xaxis.label.set_color('white')
    ax.set_xlabel('Packets')
    ax.text(0.5, 1.0, "Ratio of CSI(30x)", transform=ax.transAxes,
            ha="center", va="bottom", color="w",
            family="sans-serif", fontweight="light", fontsize=16)

    lens = cache.shape[-1]
    X = np.linspace(-int(lens/2), int(lens/2), lens)
    lines = []
    for i in range(len(cache)):
        xscale = 1 - i / 200.
        lw = 1.0 - i / 75.0
        line, = ax.plot(xscale * X, i/2 + cache[i], lw=lw, alpha=0.5)
        lines.append(line)

    def animate(i):
        global cache, mutex
        mutex.acquire()
        for i in range(len(cache)):
            lines[i].set_ydata(i/2 + cache[i])
        mutex.release()
        return lines

    ani = animation.FuncAnimation(fig, animate, interval=1000/30, blit=False)
    plt.tight_layout()
    plt.show()


if __name__ == '__main__':
    task = GetDataThread()
    task.start()
    realtime_plot()
    task.stop()
